"""Graph-operations mixin for AcherionDesigner.

Provides system-node sync, node CRUD, connection management, and all
methods that mutate the graph model. Intended to be inherited by
AcherionDesigner only.
"""

import copy
import re
import uuid
from typing import Any, cast

import acherion.node_behaviors as acherion_node_behaviors

from acherion.constants import (
    _GRID_SNAP_SIZE,
    _GROUP_COLOURS,
    _MANUAL_X,
    _MANUAL_Y,
    _NODE_STEP,
    _NODE_TOP,
    _SOURCE_X,
)
from acherion.events import (
    AcherionExternalEvent,
    default_acherion_external_events,
    normalize_acherion_external_events,
)
from acherion.registry import (
    get_acherion_node_definition,
    is_acherion_system_sink_kind,
    is_acherion_system_source_kind,
)
from acherion.model import (
    AcherionNode,
    _default_node,
    _system_node_id,
)


_AcherionLayoutMetric = dict[str, float | int]
_AcherionLayoutMetrics = dict[str, _AcherionLayoutMetric]
_AcherionLayoutBounds = dict[str, float]
_AcherionLayoutTargets = dict[str, tuple[int, int]]
_AcherionLayoutCommandResult = tuple[bool, str | _AcherionLayoutTargets]

_LAYOUT_COMMAND_SPECS: dict[str, tuple[str, object]] = {
    'align_top': ('align', ('y', 'top')),
    'align_middle': ('align', ('y', 'center')),
    'align_bottom': ('align', ('y', 'bottom')),
    'align_left': ('align', ('x', 'left')),
    'align_center': ('align', ('x', 'center')),
    'align_right': ('align', ('x', 'right')),
    'straighten_connections': ('straighten', None),
    'distribute_horizontally': ('distribute', 'x'),
    'distribute_vertically': ('distribute', 'y'),
    'stack_horizontally': ('stack', 'x'),
    'stack_vertically': ('stack', 'y'),
}

_LAYOUT_COMMAND_MESSAGES: dict[str, str] = {
    'align_top': 'Aligned selected nodes to the top edge.',
    'align_middle': 'Aligned selected nodes to the vertical middle.',
    'align_bottom': 'Aligned selected nodes to the bottom edge.',
    'align_left': 'Aligned selected nodes to the left edge.',
    'align_center': 'Aligned selected nodes to the horizontal center.',
    'align_right': 'Aligned selected nodes to the right edge.',
    'straighten_connections': 'Straightened the selected node layout.',
    'distribute_horizontally': 'Distributed selected nodes horizontally.',
    'distribute_vertically': 'Distributed selected nodes vertically.',
    'stack_horizontally': 'Stacked selected nodes horizontally.',
    'stack_vertically': 'Stacked selected nodes vertically.',
}


class _GraphOpsMixin:
    """Graph-mutation methods for AcherionDesigner."""

    _CENTER_ADD_OFFSET = 48
    _CENTER_ADD_MAX_ATTEMPTS = 64
    _COPY_PASTE_OFFSET = _GRID_SNAP_SIZE * 2
    _STACK_GAP = _GRID_SNAP_SIZE

    _selected_connection_id: str | None = None

    @staticmethod
    def _snap_grid_value(value: int | float) -> int:
        return int(round(float(value) / _GRID_SNAP_SIZE) * _GRID_SNAP_SIZE)

    def _snap_grid_point(
        self: Any,
        x: int | float,
        y: int | float,
    ) -> tuple[int, int]:
        return (
            self._snap_grid_value(x),
            self._snap_grid_value(y),
        )

    # --- node predicates ---------------------------------------------------

    @staticmethod
    def _is_system_node(node: AcherionNode) -> bool:
        return bool(node.params.get('system_node'))

    def _is_system_source_node(self: Any, node: AcherionNode) -> bool:
        return self._is_system_node(node) and is_acherion_system_source_kind(
            node.kind
        )

    def _is_system_sink_node(self: Any, node: AcherionNode) -> bool:
        return self._is_system_node(node) and is_acherion_system_sink_kind(
            node.kind
        )

    @staticmethod
    def _pure_node_id(source_id: str) -> str:
        """Strip '@pin_index' suffix from a source id."""
        return source_id.split('@')[0] if source_id and '@' in source_id else (source_id or '')

    @staticmethod
    def _source_pin_index(source_id: str) -> int:
        """Return the output pin index encoded in a source id (default 0)."""
        if source_id and '@' in source_id:
            try:
                return int(source_id.split('@', 1)[1])
            except ValueError:
                return 0
        return 0

    def _manual_nodes(self: Any) -> list[AcherionNode]:
        return [
            n for n in self._graph.nodes if not self._is_system_node(n)
        ]

    def _external_events(self: Any) -> list[AcherionExternalEvent]:
        events: dict[str, AcherionExternalEvent]
        if self._host is not None:
            events = normalize_acherion_external_events(
                self._host.external_events()
            )
        else:
            events = default_acherion_external_events()
        for node in self._manual_nodes():
            definition = get_acherion_node_definition(node.kind)
            if definition is None:
                continue
            for event in definition.external_events(self, node):
                events[event.event_key] = event
        return list(normalize_acherion_external_events(events).values())

    def _sync_manual_schema_keys(self: Any) -> None:
        if self._host is not None:
            self._host.sync_manual_schema_keys(self)

    def _source_nodes(self: Any) -> list[AcherionNode]:
        return [
            n for n in self._graph.nodes if self._is_system_source_node(n)
        ]

    def _sink_nodes(self: Any) -> list[AcherionNode]:
        return [
            n for n in self._graph.nodes if self._is_system_sink_node(n)
        ]

    def _rebuild_graph(
        self: Any,
        manual_nodes: list[AcherionNode],
    ) -> None:
        self._graph.nodes = (
            self._source_nodes() + manual_nodes + self._sink_nodes()
        )
        self._ensure_function_box_entries()
        self._sync_function_box_ports()

    # --- position seeding --------------------------------------------------

    def _take_matching_node(  # type: ignore[return]
        self,
        *,
        consumed_ids: set[str],
        kind: str,
        key: str,
        param_key: str,
    ) -> AcherionNode | None:
        owner = cast(Any, self)
        expected_id = _system_node_id(kind, key)
        for node in owner._graph.nodes:
            if node.node_id in consumed_ids:
                continue
            if node.node_id == expected_id:
                consumed_ids.add(node.node_id)
                matched_node: AcherionNode = node
                return matched_node
            if node.kind != kind:
                continue
            if str(node.params.get(param_key) or '').strip() != key:
                continue
            consumed_ids.add(node.node_id)
            matched_node = node
            return matched_node
        return None

    def _seed_position(
        self: Any,
        node: AcherionNode,
        *,
        group: str,
        index: int,
    ) -> None:
        if bool(node.params.get('manual_position')):
            return
        if group == 'source':
            node.params.setdefault('dock', 'left')
            node.params.setdefault('y', _NODE_TOP + (index * _NODE_STEP))
            return
        if group == 'sink':
            node.params.setdefault('dock', 'right')
            node.params.setdefault('y', _NODE_TOP + (index * _NODE_STEP))
            return
        node.params.setdefault('dock', 'free')
        node.params.setdefault('x', _MANUAL_X)
        node.params.setdefault('y', _MANUAL_Y + (index * _NODE_STEP))

    def _rects_overlap(
        self: Any,
        *,
        left: int,
        top: int,
        width: int,
        height: int,
        other_left: int,
        other_top: int,
        other_width: int,
        other_height: int,
        padding: int = 24,
    ) -> bool:
        return not (
            (left + width + padding) <= other_left
            or left >= (other_left + other_width + padding)
            or (top + height + padding) <= other_top
            or top >= (other_top + other_height + padding)
        )

    def _visible_node_collision(
        self: Any,
        node: AcherionNode,
        *,
        left: int,
        top: int,
    ) -> bool:
        width = self._node_width(node)
        height = self._node_height(node)
        owner = cast(Any, self)
        for other_node in owner._graph.nodes:
            if other_node.node_id == node.node_id:
                continue
            if (
                owner._is_function_input(other_node)
                or owner._is_function_output(other_node)
                or owner._is_function_entry(other_node)
            ):
                continue
            if self._rects_overlap(
                left=left,
                top=top,
                width=width,
                height=height,
                other_left=self._node_world_left(other_node),
                other_top=self._node_world_top(other_node),
                other_width=self._node_width(other_node),
                other_height=self._node_height(other_node),
            ):
                return True
        return False

    def _resolve_centered_manual_position(
        self: Any,
        node: AcherionNode,
        *,
        center_x: int,
        center_y: int,
    ) -> tuple[int, int]:
        width = self._node_width(node)
        height = self._node_height(node)
        base_left = center_x - max(1, width // 2)
        base_top = center_y - max(1, height // 2)
        for attempt in range(self._CENTER_ADD_MAX_ATTEMPTS):
            offset = attempt * self._CENTER_ADD_OFFSET
            candidate_left = base_left + offset
            candidate_top = base_top + offset
            if not self._visible_node_collision(
                node,
                left=candidate_left,
                top=candidate_top,
            ):
                return candidate_left, candidate_top
        return base_left, base_top

    # --- system node sync --------------------------------------------------

    def _sync_system_nodes(self: Any) -> None:  # noqa: C901
        if self._host is not None:
            self._host.sync_system_nodes(self)
        return

    # --- node CRUD ---------------------------------------------------------

    def _add_node(
        self: Any,
        kind: str,
        center_x: int | None = None,
        center_y: int | None = None,
    ) -> None:
        manual_nodes = self._manual_nodes()
        new_node = _default_node(kind)
        if center_x is not None and center_y is not None:
            left, top = self._resolve_centered_manual_position(
                new_node,
                center_x=center_x,
                center_y=center_y,
            )
            new_node.params['x'] = left
            new_node.params['y'] = top
            new_node.params['dock'] = 'free'
            new_node.params['manual_position'] = True
            new_node.params['parent_function'] = self._containing_function_box_id(
                center_x,
                center_y,
            )
        else:
            self._seed_position(
                new_node, group='manual', index=len(manual_nodes)
            )
            target_box_id = self._selected_function_box_target()
            if target_box_id:
                self._place_node_inside_function_box(new_node, target_box_id)
        self._ensure_custom_function_entry(new_node)
        manual_nodes.append(new_node)
        self._rebuild_graph(manual_nodes)
        self._notify_change()

    def _add_node_at_position(
        self: Any, kind: str, x: int, y: int
    ) -> None:
        """Add a node of kind at a specific canvas position from a drag-drop."""
        manual_nodes = self._manual_nodes()
        new_node = _default_node(kind)
        x, y = self._snap_grid_point(x, y)
        new_node.params['x'] = x
        new_node.params['y'] = y
        new_node.params['dock'] = 'free'
        new_node.params['manual_position'] = True
        center_x = x + max(1, self._node_width(new_node) // 2)
        center_y = y + max(1, self._node_height(new_node) // 2)
        new_node.params['parent_function'] = self._containing_function_box_id(
            center_x,
            center_y,
        )
        self._ensure_custom_function_entry(new_node)
        manual_nodes.append(new_node)
        self._rebuild_graph(manual_nodes)
        self._notify_change()

    def _copy_selection_nodes(self: Any) -> list[AcherionNode]:
        """Return manual nodes currently eligible for clipboard copy."""
        selected_ids = set(self._selected_node_ids)
        if not selected_ids:
            return []
        expanded_ids = set(selected_ids)
        for node in self._manual_nodes():
            if node.node_id not in selected_ids:
                continue
            if not self._is_function_box(node):
                continue
            expanded_ids.update(
                child.node_id
                for child in self._function_box_descendants(node.node_id)
            )
        return [
            node
            for node in self._manual_nodes()
            if node.node_id in expanded_ids
            and not self._is_function_entry(node)
            and not self._is_function_input(node)
            and not self._is_function_output(node)
        ]

    @staticmethod
    def _copy_count_message(action: str, count: int) -> str:
        """Return a short clipboard status message."""
        noun = 'node' if count == 1 else 'nodes'
        return f'{action} {count} {noun}.'

    def _copy_group_snapshot(
        self: Any,
        nodes: list[AcherionNode],
    ) -> dict[str, str]:
        """Return copied group metadata keyed by original group name."""
        snapshot: dict[str, str] = {}
        for node in nodes:
            group_name = str(node.params.get('group') or '').strip()
            if not group_name or group_name in snapshot:
                continue
            snapshot[group_name] = str(
                self._graph.groups.get(group_name) or self._next_group_colour()
            )
        return snapshot

    def _copy_user_function_snapshot(
        self: Any,
        nodes: list[AcherionNode],
    ) -> dict[str, dict[str, Any]]:
        """Return copied custom-function definitions keyed by function path."""
        snapshot: dict[str, dict[str, Any]] = {}
        for node in nodes:
            if node.kind != 'custom_function':
                continue
            function_path = str(node.params.get('function_path') or '').strip()
            if not function_path.startswith('user.'):
                continue
            if function_path in snapshot:
                continue
            snapshot[function_path] = copy.deepcopy(
                dict((self._graph.user_functions or {}).get(function_path) or {})
            )
        return snapshot

    def _copy_selection_to_clipboard(self: Any) -> tuple[bool, str]:
        """Copy the current node selection into the designer clipboard."""
        nodes = self._copy_selection_nodes()
        if not nodes:
            return (False, 'Select nodes or a function box first.')
        self._clipboard_snapshot = {
            'groups': self._copy_group_snapshot(nodes),
            'nodes': [copy.deepcopy(node) for node in nodes],
            'user_functions': self._copy_user_function_snapshot(nodes),
        }
        self._clipboard_paste_count = 0
        return (True, self._copy_count_message('Copied', len(nodes)))

    def _next_group_copy_name(
        self: Any,
        group_name: str,
        reserved_names: set[str],
    ) -> str:
        """Return a unique pasted group name derived from one group."""
        base_name = str(group_name or '').strip() or 'Group'
        candidate = f'{base_name} Copy'
        suffix = 2
        while candidate in reserved_names or candidate in self._graph.groups:
            candidate = f'{base_name} Copy {suffix}'
            suffix += 1
        reserved_names.add(candidate)
        return candidate

    def _next_pasted_function_name(
        self: Any,
        node: AcherionNode,
        reserved_names: set[str],
    ) -> str:
        """Return a unique helper name for a pasted function box."""
        base_name = self._sanitize_identifier(
            str(node.params.get('function_name') or node.title or node.node_id),
            'function_box',
        )
        candidate = f'{base_name}_copy'
        suffix = 2
        while candidate in reserved_names:
            candidate = f'{base_name}_copy_{suffix}'
            suffix += 1
        reserved_names.add(candidate)
        return candidate

    @staticmethod
    def _rename_copied_function_source(
        source_code: str,
        function_name: str,
    ) -> str:
        """Rename the single top-level function defined in copied source."""
        clean_source = str(source_code or '').rstrip()
        clean_name = str(function_name or '').strip()
        if not clean_source or not clean_name:
            return clean_source
        return re.sub(
            r'(^\s*def\s+)[A-Za-z_][A-Za-z0-9_]*(\s*\()',
            rf'\1{clean_name}\2',
            clean_source,
            count=1,
            flags=re.M,
        )

    def _duplicate_custom_function_path(
        self: Any,
        function_path: str,
        copied_user_functions: dict[str, dict[str, Any]],
    ) -> str:
        """Create a detached user-function entry for one pasted node."""
        clean_path = str(function_path or '').strip()
        if not clean_path.startswith('user.'):
            return clean_path
        function_name = self._next_custom_function_name()
        new_path = f'user.{function_name}'
        current_data = dict(
            copied_user_functions.get(clean_path)
            or (self._graph.user_functions or {}).get(clean_path)
            or {}
        )
        source_code = self._rename_copied_function_source(
            str(current_data.get('source_code') or ''),
            function_name,
        )
        if not source_code:
            source_code = self._default_custom_function_source(function_name)
        data, _error = self._parse_custom_function_source(source_code)
        if data is None:
            data = {
                'label': function_name,
                'signature': f'{function_name}()',
                'min_args': 0,
                'max_args': 0,
                'param_names': [],
                'param_types': [],
                'return_type': 'any',
                'source_code': source_code.rstrip() + '\n',
            }
        self._graph.user_functions[new_path] = data
        return new_path

    def _remap_copied_source_id(
        self: Any,
        source_id: str,
        node_id_map: dict[str, str],
    ) -> str:
        """Return the pasted source id for one copied connection."""
        clean_source_id = str(source_id or '').strip()
        if not clean_source_id:
            return ''
        pure_node_id = self._pure_node_id(clean_source_id)
        mapped_node_id = node_id_map.get(pure_node_id)
        if mapped_node_id is None:
            return ''
        if '@' not in clean_source_id:
            return mapped_node_id
        return f'{mapped_node_id}@{clean_source_id.split("@", 1)[1]}'

    def _remap_copied_node_params(
        self: Any,
        node: AcherionNode,
        node_id_map: dict[str, str],
    ) -> None:
        """Rewrite internal source references on one pasted node clone."""
        exec_sources = [
            self._remap_copied_source_id(source_id, node_id_map)
            for source_id in list(node.params.get('exec_sources') or [])
        ]
        node.params['exec_sources'] = [
            source_id for source_id in exec_sources if source_id
        ]

        if 'arg_sources' in node.params:
            node.params['arg_sources'] = [
                self._remap_copied_source_id(source_id, node_id_map)
                for source_id in list(node.params.get('arg_sources') or [])
            ]

        if 'named_sources' in node.params:
            node.params['named_sources'] = {
                param_name: mapped_source_id
                for param_name, source_id in dict(
                    node.params.get('named_sources') or {}
                ).items()
                for mapped_source_id in [
                    self._remap_copied_source_id(source_id, node_id_map)
                ]
                if mapped_source_id
            }

        if 'box_input_sources' in node.params:
            node.params['box_input_sources'] = {
                input_node_id: mapped_source_id
                for input_node_id, source_id in dict(
                    node.params.get('box_input_sources') or {}
                ).items()
                for mapped_source_id in [
                    self._remap_copied_source_id(source_id, node_id_map)
                ]
                if mapped_source_id
            }

        definition = get_acherion_node_definition(node.kind)
        if definition is None:
            return
        for param_id in definition.source_param_ids(node):
            node.params[param_id] = self._remap_copied_source_id(
                str(node.params.get(param_id) or ''),
                node_id_map,
            )

    def _next_pasted_node_id(self: Any, existing_ids: set[str]) -> str:
        """Return a unique node id for one pasted node."""
        while True:
            node_id = uuid.uuid4().hex[:8]
            if node_id in existing_ids:
                continue
            existing_ids.add(node_id)
            return node_id

    def _pasted_selection_offset(
        self: Any,
        copied_nodes: list[AcherionNode],
        *,
        anchor_x: int | None,
        anchor_y: int | None,
    ) -> tuple[int, int]:
        """Return world-space paste offset for copied nodes."""
        if anchor_x is None or anchor_y is None:
            offset = self._COPY_PASTE_OFFSET * (self._clipboard_paste_count + 1)
            return (offset, offset)
        snapped_anchor_x, snapped_anchor_y = self._snap_grid_point(
            anchor_x,
            anchor_y,
        )
        min_left = min(self._node_world_left(node) for node in copied_nodes)
        min_top = min(self._node_world_top(node) for node in copied_nodes)
        return (
            snapped_anchor_x - min_left,
            snapped_anchor_y - min_top,
        )

    def _anchor_pasted_parent_function(
        self: Any,
        node: AcherionNode,
        *,
        left: int,
        top: int,
    ) -> str:
        """Return target parent function for one cursor-anchored paste."""
        if self._is_function_box(node):
            return ''
        center_x = left + max(1, self._node_width(node) // 2)
        center_y = top + max(1, self._node_height(node) // 2)
        return self._containing_function_box_id(center_x, center_y)

    def _paste_copied_nodes(
        self: Any,
        *,
        anchor_x: int | None = None,
        anchor_y: int | None = None,
    ) -> tuple[bool, str]:
        """Paste the current clipboard nodes with rewritten references."""
        snapshot = dict(self._clipboard_snapshot or {})
        copied_nodes = [
            copy.deepcopy(node)
            for node in list(snapshot.get('nodes') or [])
            if isinstance(node, AcherionNode)
        ]
        copied_user_functions = {
            str(path): copy.deepcopy(dict(data))
            for path, data in dict(snapshot.get('user_functions') or {}).items()
            if isinstance(path, str) and isinstance(data, dict)
        }
        if not copied_nodes:
            return (False, 'Copy nodes first.')

        offset_x, offset_y = self._pasted_selection_offset(
            copied_nodes,
            anchor_x=anchor_x,
            anchor_y=anchor_y,
        )
        node_id_map: dict[str, str] = {}
        existing_ids = {node.node_id for node in self._graph.nodes}
        for node in copied_nodes:
            node_id_map[node.node_id] = self._next_pasted_node_id(existing_ids)
        for node in copied_nodes:
            if not self._is_function_box(node):
                continue
            node_id_map[self._function_entry_node_id(node.node_id)] = (
                self._function_entry_node_id(node_id_map[node.node_id])
            )

        reserved_group_names = set(self._graph.groups)
        group_name_map: dict[str, str] = {}
        for group_name, colour in dict(snapshot.get('groups') or {}).items():
            pasted_group_name = self._next_group_copy_name(
                group_name,
                reserved_group_names,
            )
            group_name_map[str(group_name)] = pasted_group_name
            self._graph.groups[pasted_group_name] = str(
                colour or self._next_group_colour()
            )

        reserved_function_names = {
            str(node.params.get('function_name') or '').strip()
            for node in self._manual_nodes()
            if self._is_function_box(node)
        }
        pasted_nodes: list[AcherionNode] = []
        for original_node in copied_nodes:
            pasted_node = copy.deepcopy(original_node)
            pasted_node.node_id = node_id_map[original_node.node_id]
            pasted_node.params = copy.deepcopy(pasted_node.params)
            left = self._node_world_left(original_node) + offset_x
            top = self._node_world_top(original_node) + offset_y
            pasted_left, pasted_top = self._snap_grid_point(left, top)
            pasted_node.params['x'] = pasted_left
            pasted_node.params['y'] = pasted_top
            pasted_node.params['dock'] = 'free'
            pasted_node.params['manual_position'] = True

            parent_function_id = str(
                original_node.params.get('parent_function') or ''
            ).strip()
            if parent_function_id:
                mapped_parent_id = node_id_map.get(parent_function_id, '')
                if mapped_parent_id:
                    pasted_node.params['parent_function'] = mapped_parent_id
                elif anchor_x is None or anchor_y is None:
                    pasted_node.params['parent_function'] = parent_function_id
                else:
                    anchored_parent_id = self._anchor_pasted_parent_function(
                        pasted_node,
                        left=pasted_left,
                        top=pasted_top,
                    )
                    if anchored_parent_id:
                        pasted_node.params['parent_function'] = (
                            anchored_parent_id
                        )
                    else:
                        pasted_node.params.pop('parent_function', None)
            else:
                if anchor_x is None or anchor_y is None:
                    pasted_node.params.pop('parent_function', None)
                else:
                    anchored_parent_id = self._anchor_pasted_parent_function(
                        pasted_node,
                        left=pasted_left,
                        top=pasted_top,
                    )
                    if anchored_parent_id:
                        pasted_node.params['parent_function'] = (
                            anchored_parent_id
                        )
                    else:
                        pasted_node.params.pop('parent_function', None)

            group_name = str(original_node.params.get('group') or '').strip()
            if group_name in group_name_map:
                pasted_node.params['group'] = group_name_map[group_name]
            else:
                pasted_node.params.pop('group', None)

            if self._is_function_box(pasted_node):
                pasted_node.params['function_name'] = (
                    self._next_pasted_function_name(
                        original_node,
                        reserved_function_names,
                    )
                )

            if pasted_node.kind == 'custom_function':
                new_path = self._duplicate_custom_function_path(
                    str(original_node.params.get('function_path') or ''),
                    copied_user_functions,
                )
                pasted_node.params['function_path'] = new_path
                pasted_node.params['module'] = self._function_path_to_module(
                    new_path
                )

            self._remap_copied_node_params(pasted_node, node_id_map)
            pasted_nodes.append(pasted_node)

        manual_nodes = self._manual_nodes()
        manual_nodes.extend(pasted_nodes)
        self._rebuild_graph(manual_nodes)
        self._selected_connection_id = None
        self._selected_node_ids = {
            node.node_id for node in pasted_nodes
        }
        self._clipboard_paste_count += 1
        self._notify_change()
        return (True, self._copy_count_message('Pasted', len(pasted_nodes)))

    def _move_node(self: Any, node_id: str, direction: int) -> None:
        manual_nodes = self._manual_nodes()
        old_index = next(
            (
                i for i, n in enumerate(manual_nodes)
                if n.node_id == node_id
            ),
            -1,
        )
        if old_index < 0:
            return
        new_index = max(0, min(len(manual_nodes) - 1, old_index + direction))
        if old_index == new_index:
            return
        node = manual_nodes.pop(old_index)
        manual_nodes.insert(new_index, node)
        self._rebuild_graph(manual_nodes)
        self._notify_change()

    def _selected_layout_nodes(self: Any) -> list[AcherionNode]:
        selected_ids = set(self._selected_node_ids)
        if not selected_ids:
            return []
        selected_nodes = [
            node for node in self._graph.nodes
            if node.node_id in selected_ids and not self._is_function_entry(node)
        ]
        selected_box_ids = {
            node.node_id
            for node in selected_nodes
            if self._is_function_box(node)
        }
        return [
            node
            for node in selected_nodes
            if not any(
                box_id != node.node_id
                and self._node_has_function_ancestor(node.node_id, box_id)
                for box_id in selected_box_ids
            )
        ]

    def _layout_metrics(
        self: Any,
        node: AcherionNode,
    ) -> _AcherionLayoutMetric:
        left = int(self._node_world_left(node))
        top = int(self._node_world_top(node))
        width = int(self._node_width(node))
        height = int(self._node_height(node))
        return {
            'left': left,
            'top': top,
            'right': left + width,
            'bottom': top + height,
            'width': width,
            'height': height,
            'center_x': left + (width / 2.0),
            'center_y': top + (height / 2.0),
        }

    def _function_box_descendants(
        self: Any,
        box_id: str,
    ) -> list[AcherionNode]:
        return [
            node
            for node in self._manual_nodes()
            if node.node_id != box_id
            and self._node_has_function_ancestor(node.node_id, box_id)
        ]

    @staticmethod
    def _median(values: list[float]) -> float:
        ordered = sorted(values)
        mid = len(ordered) // 2
        if len(ordered) % 2:
            return ordered[mid]
        return (ordered[mid - 1] + ordered[mid]) / 2.0

    def _apply_layout_positions(
        self: Any,
        nodes: list[AcherionNode],
        targets: _AcherionLayoutTargets,
    ) -> bool:
        if not targets:
            return False
        current_positions = {
            node.node_id: (
                int(self._node_world_left(node)),
                int(self._node_world_top(node)),
            )
            for node in self._graph.nodes
        }
        resolved_targets = {
            node_id: self._snap_grid_point(left, top)
            for node_id, (left, top) in targets.items()
        }
        for node in nodes:
            if not self._is_function_box(node):
                continue
            current_left, current_top = current_positions.get(
                node.node_id,
                (0, 0),
            )
            target_left, target_top = resolved_targets.get(
                node.node_id,
                (current_left, current_top),
            )
            delta_x = target_left - current_left
            delta_y = target_top - current_top
            if delta_x == 0 and delta_y == 0:
                continue
            for child in self._function_box_descendants(node.node_id):
                child_left, child_top = current_positions.get(
                    child.node_id,
                    (
                        int(self._node_world_left(child)),
                        int(self._node_world_top(child)),
                    ),
                )
                resolved_targets[child.node_id] = self._snap_grid_point(
                    child_left + delta_x,
                    child_top + delta_y,
                )

        changed = False
        moved_nodes: list[tuple[AcherionNode, int, int]] = []
        for node_id, (left, top) in resolved_targets.items():
            node = self._node_by_id(node_id)
            if node is None:
                continue
            current_left, current_top = current_positions.get(
                node_id,
                (left, top),
            )
            if current_left == left and current_top == top:
                continue
            node.params['x'] = left
            node.params['y'] = top
            node.params['dock'] = 'free'
            node.params['manual_position'] = True
            moved_nodes.append((node, left, top))
            changed = True
        if not changed:
            return False
        for moved_node, left, top in moved_nodes:
            self._assign_node_to_containing_function_box(moved_node, left, top)
        self._notify_change()
        return True

    @staticmethod
    def _layout_selection_bounds(
        metrics: _AcherionLayoutMetrics,
    ) -> _AcherionLayoutBounds:
        left = min(float(data['left']) for data in metrics.values())
        top = min(float(data['top']) for data in metrics.values())
        right = max(float(data['right']) for data in metrics.values())
        bottom = max(float(data['bottom']) for data in metrics.values())
        return {
            'left': left,
            'top': top,
            'right': right,
            'bottom': bottom,
            'center_x': (left + right) / 2.0,
            'center_y': (top + bottom) / 2.0,
        }

    @staticmethod
    def _ordered_layout_nodes(
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        *,
        primary: str,
        secondary: str,
    ) -> list[AcherionNode]:
        return sorted(
            nodes,
            key=lambda node: (
                float(metrics[node.node_id][primary]),
                float(metrics[node.node_id][secondary]),
            ),
        )

    def _layout_axis_center_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        *,
        axis: str,
        center_value: float,
    ) -> _AcherionLayoutTargets:
        targets: _AcherionLayoutTargets = {}
        for node in nodes:
            data = metrics[node.node_id]
            if axis == 'y':
                targets[node.node_id] = (
                    int(data['left']),
                    self._snap_grid_value(
                        center_value - (float(data['height']) / 2.0)
                    ),
                )
                continue
            targets[node.node_id] = (
                self._snap_grid_value(
                    center_value - (float(data['width']) / 2.0)
                ),
                int(data['top']),
            )
        return targets

    def _layout_align_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
        *,
        axis: str,
        anchor: str,
    ) -> _AcherionLayoutTargets:
        if anchor == 'center':
            return self._layout_axis_center_targets(
                nodes,
                metrics,
                axis=axis,
                center_value=float(bounds['center_y' if axis == 'y' else 'center_x']),
            )

        targets: _AcherionLayoutTargets = {}
        for node in nodes:
            data = metrics[node.node_id]
            if axis == 'y':
                anchor_values = {
                    'top': float(bounds['top']),
                    'bottom': float(bounds['bottom']) - float(data['height']),
                }
                targets[node.node_id] = (
                    int(data['left']),
                    self._snap_grid_value(anchor_values[anchor]),
                )
                continue
            anchor_values = {
                'left': float(bounds['left']),
                'right': float(bounds['right']) - float(data['width']),
            }
            targets[node.node_id] = (
                self._snap_grid_value(anchor_values[anchor]),
                int(data['top']),
            )
        return targets

    def _layout_straighten_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
    ) -> _AcherionLayoutTargets:
        horizontal = (
            (float(bounds['right']) - float(bounds['left']))
            >= (float(bounds['bottom']) - float(bounds['top']))
        )
        if horizontal:
            target_center = self._median([
                float(data['center_y']) for data in metrics.values()
            ])
            return self._layout_axis_center_targets(
                nodes,
                metrics,
                axis='y',
                center_value=target_center,
            )
        target_center = self._median([
            float(data['center_x']) for data in metrics.values()
        ])
        return self._layout_axis_center_targets(
            nodes,
            metrics,
            axis='x',
            center_value=target_center,
        )

    def _layout_distribute_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
        *,
        axis: str,
    ) -> _AcherionLayoutCommandResult:
        if len(nodes) < 3:
            if axis == 'x':
                return (
                    False,
                    'Select at least three nodes to distribute horizontally.',
                )
            return (
                False,
                'Select at least three nodes to distribute vertically.',
            )

        if axis == 'x':
            ordered = self._ordered_layout_nodes(
                nodes,
                metrics,
                primary='left',
                secondary='top',
            )
            total_size = sum(
                float(metrics[node.node_id]['width']) for node in ordered
            )
            gap = max(
                0.0,
                (float(bounds['right']) - float(bounds['left']) - total_size)
                / (len(ordered) - 1),
            )
            cursor = float(bounds['left'])
            targets: _AcherionLayoutTargets = {}
            for node in ordered:
                data = metrics[node.node_id]
                targets[node.node_id] = (
                    self._snap_grid_value(cursor),
                    int(data['top']),
                )
                cursor += float(data['width']) + gap
            return (True, targets)

        ordered = self._ordered_layout_nodes(
            nodes,
            metrics,
            primary='top',
            secondary='left',
        )
        total_size = sum(
            float(metrics[node.node_id]['height']) for node in ordered
        )
        gap = max(
            0.0,
            (float(bounds['bottom']) - float(bounds['top']) - total_size)
            / (len(ordered) - 1),
        )
        cursor = float(bounds['top'])
        targets = {}
        for node in ordered:
            data = metrics[node.node_id]
            targets[node.node_id] = (
                int(data['left']),
                self._snap_grid_value(cursor),
            )
            cursor += float(data['height']) + gap
        return (True, targets)

    def _layout_stack_targets(
        self: Any,
        nodes: list[AcherionNode],
        metrics: _AcherionLayoutMetrics,
        bounds: _AcherionLayoutBounds,
        *,
        axis: str,
    ) -> _AcherionLayoutTargets:
        if axis == 'x':
            ordered = self._ordered_layout_nodes(
                nodes,
                metrics,
                primary='left',
                secondary='top',
            )
            cursor = float(bounds['left'])
            targets: _AcherionLayoutTargets = {}
            for node in ordered:
                data = metrics[node.node_id]
                targets[node.node_id] = self._snap_grid_point(
                    cursor,
                    float(bounds['center_y']) - (float(data['height']) / 2.0),
                )
                cursor += float(data['width']) + self._STACK_GAP
            return targets

        ordered = self._ordered_layout_nodes(
            nodes,
            metrics,
            primary='top',
            secondary='left',
        )
        cursor = float(bounds['top'])
        targets = {}
        for node in ordered:
            data = metrics[node.node_id]
            targets[node.node_id] = self._snap_grid_point(
                float(bounds['center_x']) - (float(data['width']) / 2.0),
                cursor,
            )
            cursor += float(data['height']) + self._STACK_GAP
        return targets

    def _layout_selected_nodes(
        self: Any,
        command: str,
    ) -> tuple[bool, str]:
        nodes = self._selected_layout_nodes()
        if len(nodes) < 2:
            return (False, 'Select at least two nodes first.')

        metrics: _AcherionLayoutMetrics = {
            node.node_id: self._layout_metrics(node)
            for node in nodes
        }
        bounds = self._layout_selection_bounds(metrics)
        command_spec = _LAYOUT_COMMAND_SPECS.get(command)
        if command_spec is None:
            return (False, 'Unknown layout command.')

        mode, config = command_spec
        dispatchers = {
            'align': lambda raw: (
                True,
                self._layout_align_targets(
                    nodes,
                    metrics,
                    bounds,
                    axis=cast(tuple[str, str], raw)[0],
                    anchor=cast(tuple[str, str], raw)[1],
                ),
            ),
            'straighten': lambda _raw: (
                True,
                self._layout_straighten_targets(nodes, metrics, bounds),
            ),
            'distribute': lambda raw: self._layout_distribute_targets(
                nodes,
                metrics,
                bounds,
                axis=cast(str, raw),
            ),
            'stack': lambda raw: (
                True,
                self._layout_stack_targets(
                    nodes,
                    metrics,
                    bounds,
                    axis=cast(str, raw),
                ),
            ),
        }
        ok, payload = dispatchers[mode](config)
        if not ok:
            return (False, cast(str, payload))
        targets = cast(_AcherionLayoutTargets, payload)

        if not self._apply_layout_positions(nodes, targets):
            return (True, 'Selection already matches that layout.')
        return (
            True,
            _LAYOUT_COMMAND_MESSAGES.get(
                command,
                'Updated selected node layout.',
            ),
        )

    def _delete_node(self: Any, node_id: str) -> None:
        node = self._node_by_id(node_id)
        if node is not None and self._is_function_entry(node):
            return
        manual_nodes = self._manual_nodes()
        if any(node.node_id == node_id and self._is_function_box(node) for node in manual_nodes):
            child_ids = {
                node.node_id for node in manual_nodes
                if self._function_parent_id(node) == node_id
            }
            child_ids.add(node_id)
            self._delete_nodes_batch(child_ids)
            return
        manual_nodes = [
            n for n in self._manual_nodes() if n.node_id != node_id
        ]
        self._rebuild_graph(manual_nodes)
        self._cleanup_custom_function_entries()
        for node in manual_nodes:
            if self._is_function_box(node):
                self._prune_box_io_metadata(node)
        self._cleanup_empty_groups()
        self._notify_change()

    def _delete_nodes_batch(self: Any, node_ids: set[str]) -> None:
        if not node_ids:
            return
        manual_nodes = self._manual_nodes()
        expanded_ids = {
            node_id for node_id in node_ids
            if not any(
                node.node_id == node_id and self._is_function_entry(node)
                for node in manual_nodes
            )
        }
        for node in manual_nodes:
            if node.node_id in node_ids and self._is_function_box(node):
                expanded_ids.update(
                    child.node_id
                    for child in manual_nodes
                    if self._function_parent_id(child) == node.node_id
                )
        manual_nodes = [
            n for n in manual_nodes
            if n.node_id not in expanded_ids
        ]
        self._rebuild_graph(manual_nodes)
        self._cleanup_custom_function_entries()
        for node in manual_nodes:
            if self._is_function_box(node):
                self._prune_box_io_metadata(node)
        self._cleanup_empty_groups()
        self._notify_change()

    # --- group management -------------------------------------------------

    def _next_group_colour(self: Any) -> str:
        """Return next unused colour from the palette."""
        used = set(self._graph.groups.values())
        for colour in _GROUP_COLOURS:
            if colour not in used:
                return str(colour)
        return str(_GROUP_COLOURS[
            len(self._graph.groups) % len(_GROUP_COLOURS)
        ])

    def _create_group(self: Any, name: str, node_ids: set[str]) -> None:
        """Create a named group and assign selected nodes to it."""
        name = name.strip()
        if not name or name in self._graph.groups:
            return
        self._graph.groups[name] = self._next_group_colour()
        for node in self._graph.nodes:
            if node.node_id in node_ids:
                node.params['group'] = name
        self._notify_change()

    def _add_nodes_to_group(self: Any, name: str, node_ids: set[str]) -> None:
        """Move nodes into an existing group."""
        if name not in self._graph.groups:
            return
        for node in self._graph.nodes:
            if node.node_id in node_ids:
                node.params['group'] = name
        self._notify_change()

    def _remove_nodes_from_group(self: Any, node_ids: set[str]) -> None:
        """Remove nodes from their current group; prune empty groups."""
        for node in self._graph.nodes:
            if node.node_id in node_ids:
                node.params.pop('group', None)
        self._cleanup_empty_groups()
        self._notify_change()

    def _cleanup_empty_groups(self: Any) -> None:
        """Delete groups that have no member nodes."""
        members: set[str] = {
            str(n.params['group'])
            for n in self._graph.nodes
            if n.params.get('group')
        }
        self._graph.groups = {
            k: v
            for k, v in self._graph.groups.items()
            if k in members
        }

    def _extract_function_from_selection(
        self: Any,
        name: str,
        *,
        return_node_id: str | None = None,
    ) -> tuple[bool, str]:
        """Extract selected nodes into a composite function box."""
        selected_ids = set(self._selected_node_ids)
        if not selected_ids:
            return (False, 'Select nodes first.')

        selected_nodes = [
            node for node in self._manual_nodes()
            if node.node_id in selected_ids
        ]
        if len(selected_nodes) != len(selected_ids):
            return (False, 'Only manual nodes can be extracted into a function.')

        if any(self._is_function_box(node) for node in selected_nodes):
            return (False, 'Selecting a function box for extraction is not supported.')
        if any(
            self._is_function_input(node) or self._is_function_output(node)
            for node in selected_nodes
        ):
            return (False, 'Function input/output nodes cannot be re-extracted.')
        if any(self._is_function_entry(node) for node in selected_nodes):
            return (False, 'Function entry nodes cannot be re-extracted.')
        parent_ids = {
            self._function_parent_id(node) for node in selected_nodes
        }
        if len(parent_ids) > 1:
            return (
                False,
                'Select nodes from the same function scope.',
            )
        common_parent_id = next(iter(parent_ids), '')

        unsupported_kinds = {
            'for_each',
            'collect',
            'branch_route',
        }
        bad_kinds = sorted({
            node.kind
            for node in selected_nodes
            if (
                node.kind in unsupported_kinds
                or acherion_node_behaviors.function_box_unsupported(node)
            )
        })
        if bad_kinds:
            return (
                False,
                'Selection contains unsupported node kinds: '
                + ', '.join(bad_kinds),
            )

        function_slug = self._sanitize_identifier(name, 'function_box')
        manual_nodes = self._manual_nodes()
        insert_index = min(
            index
            for index, node in enumerate(manual_nodes)
            if node.node_id in selected_ids
        )
        box = _default_node('function_box')
        box.title = name.strip() or function_slug.replace('_', ' ').title()
        box.params['function_name'] = function_slug
        box.params['parent_function'] = common_parent_id
        box.params['manual_position'] = True
        box.params['dock'] = 'free'
        box.params['x'] = min(
            self._node_world_left(node) for node in selected_nodes
        ) - 48
        box.params['y'] = min(
            self._node_world_top(node) for node in selected_nodes
        ) - 72

        for node in selected_nodes:
            node.params['parent_function'] = box.node_id
            node.params.pop('group', None)

        new_manual_nodes = manual_nodes[:]
        new_manual_nodes.insert(insert_index, box)

        self._rebuild_graph(new_manual_nodes)
        self._selected_node_ids = {box.node_id}
        self._notify_change()
        return (
            True,
            f'Created function box {function_slug}. Boundary wires now define its inputs and outputs.',
        )






"""Composed helper-library builders for visual-logic compilation."""

from __future__ import annotations

from acherion.compiler.emit_nodes import (
    _emit_common_compute_node,
    _emit_function_box_node,
)
from acherion.compiler.graph import (
    _FunctionBoxGraphView,
)
from acherion.compiler.scope import (
    _EmitScope,
)
from acherion.compiler.state import (
    EmitState,
    ExternalInput,
)
from acherion.compiler.utils import (
    _add_missing_pass,
    _safe_function_name,
)
from acherion.model import (
    AcherionGraph,
    AcherionNode,
)
import acherion.node_behaviors as acherion_node_behaviors


def _unique_param_name(label: str, fallback: str, used: set[str]) -> str:
    """Return a unique helper parameter name."""
    base_name = _safe_function_name(label, fallback)
    name = base_name
    suffix = 2
    while name in used:
        name = f'{base_name}_{suffix}'
        suffix += 1
    used.add(name)
    return name


def _function_entry_source_id(box_node_id: str) -> str:
    """Return canonical hidden entry source id for one function box."""
    return f'{box_node_id}:entry'


class _FunctionBodyCompiler:
    """Compile one function-box body using composed emit state."""

    def __init__(
        self,
        helper_name: str,
        member_nodes: list[AcherionNode],
        *,
        external_inputs: list[ExternalInput],
        return_source_ids: list[str],
        all_nodes: list[AcherionNode] | None = None,
    ) -> None:
        self._helper_name = helper_name
        self._member_nodes = list(member_nodes)
        self._external_inputs = list(external_inputs)
        self._return_source_ids = list(return_source_ids)
        self._scope_nodes = list(all_nodes or member_nodes)
        self._original_index = {
            node.node_id: index for index, node in enumerate(self._scope_nodes)
        }
        self._state = EmitState()
        self._function_graph = _FunctionBoxGraphView(self._scope_nodes)
        self._parent_function_id = self._resolve_parent_function_id()
        self._input_nodes = self._resolve_input_nodes()
        self._entry_node_id = self._resolve_entry_node_id()
        self._scope = _EmitScope(
            self._member_nodes,
            state=self._state,
            function_graph=self._function_graph,
            original_index=self._original_index,
            entry_node_id=self._entry_node_id,
            owner_name=None,
        )

    def compile(self) -> str:
        """Return the compiled helper source."""
        param_names = self._build_param_names()
        signature = ', '.join(param_names)
        self._state.lines.extend([
            f'def {self._helper_name}({signature}):',
        ])
        self._seed_inputs(param_names)
        self._emit_body()
        self._emit_return()
        return '\n'.join(_add_missing_pass(self._state.lines)) + '\n'

    def _resolve_parent_function_id(self) -> str:
        for node in self._member_nodes:
            parent_id = str(node.params.get('parent_function') or '')
            if parent_id:
                return parent_id
        return ''

    def _resolve_input_nodes(self) -> list[AcherionNode]:
        if not self._parent_function_id:
            return []
        parent_box = self._function_graph.node_index.get(
            self._parent_function_id,
        )
        if parent_box is None:
            return []
        return self._function_graph.ordered_io_nodes(
            self._parent_function_id,
            io_kind='function_input',
            ordered_ids=list(parent_box.params.get('input_order') or []),
        )

    def _resolve_entry_node_id(self) -> str:
        if not self._parent_function_id:
            return ''
        for node in self._member_nodes:
            if node.kind != 'function_entry':
                continue
            if str(node.params.get('parent_function') or '') != self._parent_function_id:
                continue
            return node.node_id
        return _function_entry_source_id(self._parent_function_id)

    def _build_param_names(self) -> list[str]:
        used_param_names: set[str] = set()
        param_names: list[str] = []
        external_input_names = {
            str(target_id or ''): str(name or '')
            for name, target_id in self._external_inputs
        }
        if self._input_nodes:
            for index, input_node in enumerate(self._input_nodes):
                label = external_input_names.get(input_node.node_id) or str(
                    input_node.params.get('label')
                    or input_node.params.get('param_name')
                    or input_node.title
                    or ''
                )
                param_names.append(
                    _unique_param_name(
                        label,
                        f'arg_{index + 1}',
                        used_param_names,
                    )
                )
            return param_names

        for index, (name, _source_id) in enumerate(self._external_inputs):
            fallback = f'arg_{index + 1}'
            param_names.append(
                _unique_param_name(
                    str(name or fallback),
                    fallback,
                    used_param_names,
                )
            )
        return param_names

    def _seed_inputs(self, param_names: list[str]) -> None:
        if self._input_nodes:
            for index, input_node in enumerate(self._input_nodes):
                self._state.store(input_node.node_id, param_names[index])
            return

        for (_name, source_id), param_name in zip(
            self._external_inputs,
            param_names,
        ):
            self._state.store_source(str(source_id or ''), param_name)

    def _emit_body(self) -> None:
        self._scope.emit(self._emit_helper_node)

    def _emit_helper_node(
        self,
        node: AcherionNode,
        params: dict[str, object],
        var_name: str,
        indent: str,
    ) -> None:
        kind = node.kind
        if (
            str(params.get('parent_function') or '')
            != self._parent_function_id
        ):
            return
        if kind in {
            'function_entry',
            'function_input',
            'function_output',
            'branch_route',
            'for_each',
        }:
            return

        if acherion_node_behaviors.emit_static_node(
            state=self._state,
            node=node,
            params=params,
            indent=indent,
            var_name=var_name,
            owner_name=None,
        ):
            return

        if _emit_function_box_node(
            state=self._state,
            node=node,
            params=params,
            function_graph=self._function_graph,
            indent=indent,
            var_name=var_name,
            runtime_self_name=None,
            helper_as_attribute=False,
        ):
            return

        if _emit_common_compute_node(
            state=self._state,
            node=node,
            params=params,
            indent=indent,
            var_name=var_name,
            method_owner_name=None,
        ):
            return

        if acherion_node_behaviors.function_box_unsupported(node):
            self._state.lines.append(
                f'{indent}# Unsupported inside Function Box: {kind}'
            )
            return

        self._state.lines.append(
            f'{indent}{var_name} = None '
            f'# Unsupported extracted node kind: {kind}'
        )
        self._state.store(node.node_id, var_name)

    def _emit_return(self) -> None:
        for source_id in self._return_source_ids:
            if not source_id:
                continue
            self._scope.ensure_source(
                source_id,
                indent_level=1,
                emit_node=self._emit_helper_node,
            )
        return_exprs = [
            self._state.source_expr(source_id)
            for source_id in self._return_source_ids
            if source_id
        ]
        if not return_exprs:
            self._state.lines.append('    return None')
        elif len(return_exprs) == 1:
            self._state.lines.append(f'    return {return_exprs[0]}')
        else:
            self._state.lines.append(
                f"    return ({', '.join(return_exprs)})"
            )


class _HelperLibraryBuilder:
    """Build graph-local static helper methods in dependency-safe order."""

    def __init__(
        self,
        graph: AcherionGraph,
        *,
        function_graph: _FunctionBoxGraphView,
    ) -> None:
        self._graph = graph
        self._function_graph = function_graph

    def build_lines(self) -> list[str]:
        """Return helper method lines ready to prepend before run()."""
        graph_order = {
            node.node_id: index for index, node in enumerate(self._graph.nodes)
        }
        function_boxes = [
            node for node in self._graph.nodes if node.kind == 'function_box'
        ]
        function_boxes.sort(
            key=lambda box: (
                -self._function_graph.function_box_depth(box),
                graph_order.get(box.node_id, 0),
            )
        )

        lines: list[str] = []
        for box in function_boxes:
            helper_code = self._compile_box_helper(box).strip('\n')
            if not helper_code:
                continue
            lines.extend(helper_code.splitlines())
            lines.append('')
        return lines

    def _compile_box_helper(self, box: AcherionNode) -> str:
        member_nodes = [
            node
            for node in self._graph.nodes
            if str(node.params.get('parent_function') or '') == box.node_id
        ]
        helper_name = _safe_function_name(
            str(
                box.params.get('function_name')
                or box.title
                or box.node_id
            ),
            f'function_{box.node_id}',
        )
        return _FunctionBodyCompiler(
            helper_name,
            member_nodes,
            external_inputs=self._function_graph.external_inputs(box),
            return_source_ids=self._function_graph.return_sources(box),
            all_nodes=self._graph.nodes,
        ).compile()


"""Built-in composite node definitions for Acherion."""

from __future__ import annotations

import acherion.node as acherion_node


class CustomFunctionNode(acherion_node.CompositeNodeDefinition):
    kind = 'custom_function'
    label = 'Custom Function'
    icon = 'functions'
    tooltip = 'Define and call a user method from a compact code editor dialog.'
    manual_add = True
    producer = True
    exec_in = True
    exec_out = True
    default_params_factory = acherion_node.custom_function_params


class FunctionEntryNode(acherion_node.CompositeNodeDefinition):
    kind = 'function_entry'
    label = 'Entry'
    icon = 'login'
    tooltip = 'Auto-created entry point for execution inside a Function Box.'
    exec_out = True
    default_params_factory = acherion_node.literal_params({
        'parent_function': '',
        'group': '',
    })


class FunctionBoxNode(acherion_node.CompositeNodeDefinition):
    kind = 'function_box'
    label = 'Function Box'
    icon = 'functions'
    tooltip = 'Composite function container with a compact internal entry exec pin.'
    manual_add = True
    producer = True
    exec_in = True
    exec_out = True
    default_params_factory = acherion_node.literal_params({
        'function_name': '',
    })

    def input_pins(
        self,
        owner: object,
        node: object,
    ) -> list[dict[str, str]] | None:
        del owner, node
        return []

    def output_pins(
        self,
        owner: object,
        node: object,
    ) -> list[dict[str, str]] | None:
        del owner, node
        return []


BUILTIN_COMPOSITE_NODES = (
    CustomFunctionNode(),
    FunctionEntryNode(),
    FunctionBoxNode(),
)

__all__ = ['BUILTIN_COMPOSITE_NODES']



"""Typed host contracts for embedded Acherion designers."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Protocol

from acherion.events import AcherionExternalEvent, default_acherion_external_events


AcherionRuntimeBindings = dict[str, dict[str, dict[str, Any]]]
AcherionExternalEventsHook = Callable[[], dict[str, AcherionExternalEvent]]
AcherionCodegenHook = Callable[[Any], str]
AcherionRuntimeBindingsHook = Callable[[Any], AcherionRuntimeBindings]
AcherionGraphSyncHook = Callable[[Any], None]
AcherionConfigRenderHook = Callable[..., bool]


def _default_external_events() -> dict[str, AcherionExternalEvent]:
    """Return default Acherion external events for one host."""
    return default_acherion_external_events()


def _default_runtime_bindings(designer: Any) -> AcherionRuntimeBindings:
    """Return empty runtime bindings for hosts that do not need them."""
    del designer
    return {}


def _noop_graph_sync(designer: Any) -> None:
    """Perform no graph sync for hosts that do not need it."""
    del designer


def _noop_config_render(
    designer: Any,
    node: Any,
    *,
    refresh_editor: Callable[[], None] | None,
    apply_change: Callable[..., None],
) -> bool:
    """Render no host-owned config UI by default."""
    del designer, node, refresh_editor, apply_change
    return False


class AcherionCodegenHost(Protocol):
    """Compile graph state into host-owned runtime artifacts."""

    def external_events(self) -> dict[str, AcherionExternalEvent]:
        """Return host-owned external events visible inside the designer."""

    def generated_user_code(self, designer: Any) -> str:
        """Compile the current graph into user code."""

    def generated_runtime_bindings(
        self,
        designer: Any,
    ) -> AcherionRuntimeBindings:
        """Collect runtime bindings for generated graph code."""


class AcherionGraphSyncHost(Protocol):
    """Normalize host-owned schema and system-node state."""

    def sync_manual_schema_keys(self, designer: Any) -> None:
        """Assign schema keys to manual nodes before sync."""

    def sync_system_nodes(self, designer: Any) -> None:
        """Rebuild system-owned nodes around the current manual graph."""


class AcherionConfigRenderHost(Protocol):
    """Render host-owned editor controls for configured nodes."""

    def render_system_sink_config_fields(
        self,
        designer: Any,
        node: Any,
        *,
        refresh_editor: Callable[[], None] | None,
        apply_change: Callable[..., None],
    ) -> bool:
        """Render host-owned config UI for system sink nodes."""

    def render_node_config_fields(
        self,
        designer: Any,
        node: Any,
        *,
        refresh_editor: Callable[[], None] | None,
        apply_change: Callable[..., None],
    ) -> bool:
        """Render host-owned config UI for manual nodes."""


class AcherionHost(
    AcherionCodegenHost,
    AcherionGraphSyncHost,
    AcherionConfigRenderHost,
    Protocol,
):
    """Compose full host behavior from smaller host responsibilities."""


@dataclass(slots=True)
class ComposedAcherionHost:
    """Concrete host object built from small host callback hooks."""

    _generated_user_code: AcherionCodegenHook
    _external_events: AcherionExternalEventsHook = _default_external_events
    _generated_runtime_bindings: AcherionRuntimeBindingsHook = (
        _default_runtime_bindings
    )
    _sync_manual_schema_keys: AcherionGraphSyncHook = _noop_graph_sync
    _sync_system_nodes: AcherionGraphSyncHook = _noop_graph_sync
    _render_system_sink_config_fields: AcherionConfigRenderHook = (
        _noop_config_render
    )
    _render_node_config_fields: AcherionConfigRenderHook = _noop_config_render

    def external_events(self) -> dict[str, AcherionExternalEvent]:
        """Return host-owned external events visible inside the designer."""
        return dict(self._external_events())

    def generated_user_code(self, designer: Any) -> str:
        """Compile the current graph into user code."""
        return str(self._generated_user_code(designer))

    def generated_runtime_bindings(
        self,
        designer: Any,
    ) -> AcherionRuntimeBindings:
        """Collect runtime bindings for generated graph code."""
        return dict(self._generated_runtime_bindings(designer))

    def sync_manual_schema_keys(self, designer: Any) -> None:
        """Assign schema keys to manual nodes before sync."""
        self._sync_manual_schema_keys(designer)

    def sync_system_nodes(self, designer: Any) -> None:
        """Rebuild system-owned nodes around the current manual graph."""
        self._sync_system_nodes(designer)

    def render_system_sink_config_fields(
        self,
        designer: Any,
        node: Any,
        *,
        refresh_editor: Callable[[], None] | None,
        apply_change: Callable[..., None],
    ) -> bool:
        """Render host-owned config UI for system sink nodes."""
        return bool(
            self._render_system_sink_config_fields(
                designer,
                node,
                refresh_editor=refresh_editor,
                apply_change=apply_change,
            )
        )

    def render_node_config_fields(
        self,
        designer: Any,
        node: Any,
        *,
        refresh_editor: Callable[[], None] | None,
        apply_change: Callable[..., None],
    ) -> bool:
        """Render host-owned config UI for manual nodes."""
        return bool(
            self._render_node_config_fields(
                designer,
                node,
                refresh_editor=refresh_editor,
                apply_change=apply_change,
            )
        )


def compose_acherion_host(
    *,
    generated_user_code: AcherionCodegenHook,
    external_events: AcherionExternalEventsHook | None = None,
    generated_runtime_bindings: AcherionRuntimeBindingsHook | None = None,
    sync_manual_schema_keys: AcherionGraphSyncHook | None = None,
    sync_system_nodes: AcherionGraphSyncHook | None = None,
    render_system_sink_config_fields: AcherionConfigRenderHook | None = None,
    render_node_config_fields: AcherionConfigRenderHook | None = None,
) -> ComposedAcherionHost:
    """Compose one embeddable Acherion host from small callback hooks."""
    return ComposedAcherionHost(
        _generated_user_code=generated_user_code,
        _external_events=external_events or _default_external_events,
        _generated_runtime_bindings=(
            generated_runtime_bindings or _default_runtime_bindings
        ),
        _sync_manual_schema_keys=sync_manual_schema_keys or _noop_graph_sync,
        _sync_system_nodes=sync_system_nodes or _noop_graph_sync,
        _render_system_sink_config_fields=(
            render_system_sink_config_fields or _noop_config_render
        ),
        _render_node_config_fields=(
            render_node_config_fields or _noop_config_render
        ),
    )


__all__ = [
    'AcherionCodegenHost',
    'AcherionConfigRenderHost',
    'AcherionGraphSyncHost',
    'AcherionHost',
    'AcherionRuntimeBindings',
    'ComposedAcherionHost',
    'compose_acherion_host',
]


"""NiceGUI embedding helpers for standalone or hosted Acherion."""

from __future__ import annotations

from acherion.embed.code_editor import (
    build_python_code_editor,
    inject_python_autocomplete,
)
from acherion.embed.designer.component import (
    AcherionDesigner,
)
from acherion.embed.workbench import (
    AcherionWorkbench,
)
from acherion.embed.schema_fields import (
    normalise_event_emitters,
    register_schema_component_field_renderer,
    register_schema_component_field_renderers,
    render_event_emitter_fields,
    render_schema_component_fields,
    schema_label,
)

__all__ = [
    'AcherionDesigner',
    'AcherionWorkbench',
    'build_python_code_editor',
    'inject_python_autocomplete',
    'normalise_event_emitters',
    'register_schema_component_field_renderer',
    'register_schema_component_field_renderers',
    'render_event_emitter_fields',
    'render_schema_component_fields',
    'schema_label',
]


"""Generic Python completion endpoints for standalone Acherion."""

# pyright: reportMissingTypeStubs=false

from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from typing import Any

import jedi  # pyright: ignore[reportMissingTypeStubs]
from fastapi import Request
from nicegui import app

logger = logging.getLogger(__name__)

_JEDI_PREFIX = """\
import collections
import logging
import pathlib
import re
import typing

logger = logging.getLogger(__name__)
"""

_PREFIX_LINES: int = len(_JEDI_PREFIX.splitlines())
_CLASS_SEP = "# --- class methods ---"
_JEDI_PROJECT = jedi.Project(path=str(Path(__file__).parent.parent.parent))


def _build_jedi_context(code: str, row: int, col: int) -> tuple[str, int, int]:
    """Wrap editor code in a synthetic Jedi document."""
    normalized_code = str(code or '').replace(_CLASS_SEP, '').strip('\n')
    if not normalized_code.strip():
        normalized_code = 'pass'
    full_code = _JEDI_PREFIX + normalized_code
    jedi_row = _PREFIX_LINES + row + 1
    jedi_col = col
    return full_code, jedi_row, jedi_col


def _parse_body(body: Any) -> tuple[str, int, int] | None:
    """Extract code, row, and column from a request body."""
    if not isinstance(body, dict):
        return None
    try:
        code = str(body.get('code', ''))
        row = max(0, int(body.get('row', 0)))
        col = max(0, int(body.get('col', 0)))
    except (TypeError, ValueError):
        return None
    return code, row, col


@app.post('/api/py-completions')
async def py_completions(request: Request) -> dict[str, Any]:
    """Return Jedi completions for the current editor cursor position."""
    try:
        params = _parse_body(await request.json())
    except Exception:  # Top-level request boundary.
        return {'completions': []}
    if params is None:
        return {'completions': []}
    code, row, col = params
    full_code, jedi_row, jedi_col = _build_jedi_context(code, row, col)

    def _run() -> list[dict[str, str]]:
        script = jedi.Script(full_code, project=_JEDI_PROJECT)
        raw = script.complete(jedi_row, jedi_col)
        return [
            {
                'name': completion.name,
                'type': completion.type or 'variable',
                'description': (
                    completion.description or ''
                ).split('\n')[0][:120],
            }
            for completion in raw[:80]
            if completion.name and not completion.name.startswith('_')
        ]

    try:
        loop = asyncio.get_running_loop()
        return {'completions': await loop.run_in_executor(None, _run)}
    except Exception:  # Top-level execution boundary.
        logger.exception('Jedi completion failed')
        return {'completions': []}


@app.post('/api/py-hover')
async def py_hover(request: Request) -> dict[str, Any]:
    """Return hover documentation for the symbol at the cursor."""
    try:
        params = _parse_body(await request.json())
    except Exception:  # Top-level request boundary.
        return {'docs': ''}
    if params is None:
        return {'docs': ''}
    code, row, col = params
    full_code, jedi_row, jedi_col = _build_jedi_context(code, row, col)

    def _run() -> str:
        script = jedi.Script(full_code, project=_JEDI_PROJECT)
        results = script.help(jedi_row, jedi_col)
        if not results:
            results = script.infer(jedi_row, jedi_col)
        if not results:
            return ''
        result = results[0]
        parts: list[str] = []
        description = getattr(result, 'description', None) or ''
        if description:
            parts.append(description)
        doc_fn = getattr(result, 'docstring', None)
        if callable(doc_fn):
            doc_result = doc_fn()
            doc = doc_result if isinstance(doc_result, str) else str(doc_result or '')
            if doc:
                trimmed = doc[:900]
                if len(doc) > 900:
                    trimmed += '\n...'
                parts.append(trimmed)
        return '\n\n'.join(parts)

    try:
        loop = asyncio.get_running_loop()
        return {'docs': await loop.run_in_executor(None, _run)}
    except Exception:  # Top-level execution boundary.
        logger.exception('Jedi hover failed')
        return {'docs': ''}


@app.post('/api/py-signatures')
async def py_signatures(request: Request) -> dict[str, Any]:
    """Return function signature help for the current cursor position."""
    try:
        params = _parse_body(await request.json())
    except Exception:  # Top-level request boundary.
        return {'signatures': []}
    if params is None:
        return {'signatures': []}
    code, row, col = params
    full_code, jedi_row, jedi_col = _build_jedi_context(code, row, col)

    def _run() -> list[dict[str, Any]]:
        script = jedi.Script(full_code, project=_JEDI_PROJECT)
        signatures = script.get_signatures(jedi_row, jedi_col)
        result: list[dict[str, Any]] = []
        for signature in signatures[:3]:
            param_labels = [parameter.description for parameter in signature.params]
            result.append({
                'label': signature.name + '(' + ', '.join(param_labels) + ')',
                'name': signature.name,
                'params': param_labels,
                'active': signature.index if signature.index is not None else -1,
            })
        return result

    try:
        loop = asyncio.get_running_loop()
        return {'signatures': await loop.run_in_executor(None, _run)}
    except Exception:  # Top-level execution boundary.
        logger.exception('Jedi signatures failed')
        return {'signatures': []}



"""Module discovery helpers for the visual-logic function catalog."""

from __future__ import annotations

import builtins as _builtins
import dataclasses
import functools
import importlib
import logging
import math as _math
from collections.abc import Callable, Iterable
from typing import Any

_log = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class CatalogScanSpec:
    """One scan target inside a catalog module."""

    prefix: str
    attr_path: tuple[str, ...] = ()


@dataclasses.dataclass(frozen=True)
class CatalogModuleSpec:
    """Metadata for one module exposed in the visual-logic catalog."""

    key: str
    label: str
    import_path: str | None = None
    scans: tuple[CatalogScanSpec, ...] = ()
    path_prefixes: tuple[str, ...] = ()
    loader: Callable[[], Any] | None = None
    blacklist: frozenset[str] = frozenset()
    type_allowlist: frozenset[str] = frozenset()
    object_module_roots: tuple[str, ...] = ()
    gate_types_with_allowlist: bool = False


@dataclasses.dataclass(frozen=True)
class LoadedCatalogModule:
    """A discovered module paired with its catalog metadata."""

    spec: CatalogModuleSpec
    module: Any


GLOBAL_BLACKLIST: frozenset[str] = frozenset({
    'print',
    'input',
    'open',
    'exec',
    'eval',
    'compile',
    '__import__',
    'breakpoint',
    'exit',
    'quit',
    'dir',
    'vars',
    'locals',
    'globals',
    'help',
    'id',
    'callable',
    'repr',
    'ascii',
    'format',
    'memoryview',
    'copyright',
    'credits',
    'license',
})

_BUILTINS_BLACKLIST: frozenset[str] = frozenset({
    'object', 'super', 'property', 'classmethod', 'staticmethod',
    'getattr', 'setattr', 'delattr', 'hasattr',
    'iter', 'next', 'hash',
    'bytearray', 'bytes', 'complex', 'frozenset', 'memoryview', 'slice',
    'type', '__build_class__', '__loader__', '__spec__',
    '__doc__', '__package__', '__name__',
    'bin', 'oct', 'hex', 'chr', 'ord',
    'aiter', 'anext',
})

_NUMPY_BLACKLIST: frozenset[str] = frozenset({
    'deprecate',
    'deprecate_with_doc',
    'show_config',
    'get_include',
    'who',
    'lookfor',
    'source',
    'info',
    'show_runtime',
    'dtype',
    'test',
})

_COLLECTIONS_BLACKLIST: frozenset[str] = frozenset({
    'Awaitable', 'Callable', 'Container', 'Coroutine', 'Generator',
    'Hashable', 'ItemsView', 'Iterable', 'Iterator', 'KeysView',
    'Mapping', 'MappingView', 'MutableMapping', 'MutableSequence',
    'MutableSet', 'Reversible', 'Sequence', 'Set', 'Sized', 'ValuesView',
})

_PATHLIB_BLACKLIST: frozenset[str] = frozenset({
    'PurePath',
    'PurePosixPath',
    'PureWindowsPath',
    'PosixPath',
    'WindowsPath',
})

_MATPLOTLIB_BLACKLIST: frozenset[str] = frozenset({
    'test',
    'use',
    'get_backend',
    'interactive',
    'is_interactive',
    'rc_context',
    'rcdefaults',
    'rcParams',
})

_MATPLOTLIB_PYPLOT_BLACKLIST: frozenset[str] = frozenset({
    'show',
    'ion',
    'ioff',
    'pause',
    'waitforbuttonpress',
    'switch_backend',
    'get_backend',
})

_BUILTIN_TYPE_ALLOWLIST: frozenset[str] = frozenset({
    'bool', 'dict', 'enumerate', 'filter', 'float', 'int', 'list', 'map',
    'range', 'reversed', 'set', 'str', 'tuple', 'zip',
})

_COLLECTIONS_TYPE_ALLOWLIST: frozenset[str] = frozenset({
    'Counter', 'defaultdict', 'deque', 'OrderedDict', 'ChainMap',
})

_PATHLIB_TYPE_ALLOWLIST: frozenset[str] = frozenset({'Path'})

_NUMPY_TYPE_ALLOWLIST: frozenset[str] = frozenset({'ndarray'})

_BASE_CATALOG_MODULE_SPECS: tuple[CatalogModuleSpec, ...] = (
    CatalogModuleSpec(
        key='builtins',
        label='Python builtins',
        loader=lambda: _builtins,
        scans=(CatalogScanSpec(prefix=''),),
        blacklist=_BUILTINS_BLACKLIST,
        type_allowlist=_BUILTIN_TYPE_ALLOWLIST,
        gate_types_with_allowlist=True,
    ),
    CatalogModuleSpec(
        key='math',
        label='math',
        loader=lambda: _math,
        scans=(CatalogScanSpec(prefix='math.'),),
        path_prefixes=('math.',),
        gate_types_with_allowlist=True,
    ),
    CatalogModuleSpec(
        key='re',
        label='re (regex)',
        import_path='re',
        scans=(CatalogScanSpec(prefix='re.'),),
        path_prefixes=('re.',),
        blacklist=frozenset({'purge', 'template'}),
        gate_types_with_allowlist=True,
    ),
    CatalogModuleSpec(
        key='collections',
        label='collections',
        import_path='collections',
        scans=(CatalogScanSpec(prefix='collections.'),),
        path_prefixes=('collections.',),
        blacklist=_COLLECTIONS_BLACKLIST,
        type_allowlist=_COLLECTIONS_TYPE_ALLOWLIST,
        object_module_roots=('collections',),
        gate_types_with_allowlist=True,
    ),
    CatalogModuleSpec(
        key='pathlib',
        label='pathlib',
        import_path='pathlib',
        scans=(CatalogScanSpec(prefix='pathlib.'),),
        path_prefixes=('pathlib.',),
        blacklist=_PATHLIB_BLACKLIST,
        type_allowlist=_PATHLIB_TYPE_ALLOWLIST,
        object_module_roots=('pathlib',),
        gate_types_with_allowlist=True,
    ),
    CatalogModuleSpec(
        key='np',
        label='numpy (np)',
        import_path='numpy',
        scans=(CatalogScanSpec(prefix='np.'),),
        path_prefixes=('np.', 'numpy.'),
        blacklist=_NUMPY_BLACKLIST,
        type_allowlist=_NUMPY_TYPE_ALLOWLIST,
        object_module_roots=('numpy',),
        gate_types_with_allowlist=True,
    ),
    CatalogModuleSpec(
        key='go',
        label='plotly.graph_objects (go)',
        import_path='plotly.graph_objects',
        scans=(CatalogScanSpec(prefix='go.'),),
        path_prefixes=('go.', 'plotly.graph_objects.', 'plotly.graph_objs.'),
        object_module_roots=('plotly.graph_objects', 'plotly.graph_objs'),
    ),
    CatalogModuleSpec(
        key='px',
        label='plotly.express (px)',
        import_path='plotly.express',
        scans=(CatalogScanSpec(prefix='px.'),),
        path_prefixes=('px.', 'plotly.express.'),
        object_module_roots=('plotly.express',),
        gate_types_with_allowlist=True,
    ),
    CatalogModuleSpec(
        key='plotly.subplots',
        label='plotly.subplots',
        import_path='plotly.subplots',
        scans=(CatalogScanSpec(prefix='plotly.subplots.'),),
        path_prefixes=('plotly.subplots.',),
        object_module_roots=('plotly.subplots',),
        gate_types_with_allowlist=True,
    ),
)

_REGISTERED_CATALOG_MODULE_SPECS: dict[str, CatalogModuleSpec] = {}


def _invalidate_catalog_caches() -> None:
    """Clear catalog module/runtime caches after extension changes."""
    available_modules.cache_clear()
    _module_labels.cache_clear()
    _spec_by_key.cache_clear()
    _path_prefix_to_module.cache_clear()
    runtime_global_bindings.cache_clear()
    from acherion.catalog import runtime as _catalog_runtime

    _catalog_runtime.clear_catalog_runtime_caches()


def register_catalog_module_specs(
    specs: Iterable[CatalogModuleSpec],
    *,
    replace: bool = False,
) -> None:
    """Register host-provided catalog module specs onto Acherion."""
    base_keys = {spec.key for spec in _BASE_CATALOG_MODULE_SPECS}
    for spec in specs:
        clean_key = str(spec.key or '').strip()
        if not clean_key:
            raise ValueError('Catalog module key is required.')
        if (
            clean_key in _REGISTERED_CATALOG_MODULE_SPECS
            or clean_key in base_keys
        ) and not replace:
            raise ValueError(f'Catalog module already registered: {clean_key}')
        _REGISTERED_CATALOG_MODULE_SPECS[clean_key] = spec
    _invalidate_catalog_caches()


def catalog_module_specs() -> tuple[CatalogModuleSpec, ...]:
    """Return core plus host-registered catalog module specs."""
    specs_by_key: dict[str, CatalogModuleSpec] = {
        spec.key: spec for spec in _BASE_CATALOG_MODULE_SPECS
    }
    specs_by_key.update(_REGISTERED_CATALOG_MODULE_SPECS)
    return tuple(specs_by_key.values())


@functools.lru_cache(maxsize=1)
def _module_labels() -> dict[str, str]:
    """Return cached module labels keyed by catalog module key."""
    return {spec.key: spec.label for spec in catalog_module_specs()}


@functools.lru_cache(maxsize=1)
def _spec_by_key() -> dict[str, CatalogModuleSpec]:
    """Return cached catalog specs keyed by module key."""
    return {spec.key: spec for spec in catalog_module_specs()}


@functools.lru_cache(maxsize=1)
def _path_prefix_to_module() -> tuple[tuple[str, str], ...]:
    """Return cached path-prefix map for module inference."""
    return tuple(
        sorted(
            (
                (prefix, spec.key)
                for spec in catalog_module_specs()
                for prefix in spec.path_prefixes
            ),
            key=lambda item: len(item[0]),
            reverse=True,
        )
    )


def _load_catalog_module(spec: CatalogModuleSpec) -> Any | None:
    """Load one catalog module object from import path or loader."""
    loader = spec.loader
    try:
        module_obj = (
            loader()
            if loader is not None
            else importlib.import_module(str(spec.import_path))
        )
    except ImportError:
        if spec.import_path:
            _log.debug('%s not available; omitted from catalog', spec.import_path)
        return None
    return module_obj


@functools.lru_cache(maxsize=1)
def available_modules() -> tuple[LoadedCatalogModule, ...]:
    """Return loaded catalog modules discovered from registered specs."""
    loaded: list[LoadedCatalogModule] = []
    for spec in catalog_module_specs():
        module_obj = _load_catalog_module(spec)
        if module_obj is None:
            continue
        loaded.append(LoadedCatalogModule(spec=spec, module=module_obj))
    return tuple(loaded)


def resolve_attr_path(root: Any, attr_path: str | tuple[str, ...]) -> Any | None:
    """Resolve dotted or segmented attribute paths from a root object."""
    if not attr_path:
        return root
    if isinstance(attr_path, str):
        parts = tuple(part for part in attr_path.split('.') if part)
    else:
        parts = attr_path
    current = root
    for part in parts:
        try:
            current = getattr(current, part)
        except AttributeError:
            return None
    return current


def scan_targets(
    loaded_module: LoadedCatalogModule,
) -> tuple[tuple[Any, str], ...]:
    """Return module objects and path prefixes to scan for one module."""
    targets: list[tuple[Any, str]] = []
    for scan in loaded_module.spec.scans:
        target = resolve_attr_path(loaded_module.module, scan.attr_path)
        if target is None:
            continue
        targets.append((target, scan.prefix))
    return tuple(targets)


def module_options() -> dict[str, str]:
    """Return {module_key: display_label} for the module selector."""
    return dict(_module_labels())


@functools.lru_cache(maxsize=1)
def runtime_global_bindings() -> dict[str, Any]:
    """Return runtime globals derived from registered catalog modules."""
    bindings: dict[str, Any] = {}
    for loaded_module in available_modules():
        spec = loaded_module.spec
        bindings.setdefault(spec.key, loaded_module.module)
        if spec.import_path:
            root_name = str(spec.import_path).split('.', 1)[0]
            try:
                bindings.setdefault(
                    root_name,
                    importlib.import_module(root_name),
                )
            except ImportError:
                continue
    return bindings


def path_to_module(path: str) -> str:
    """Infer module key from a function or class path."""
    for prefix, module_key in _path_prefix_to_module():
        if path.startswith(prefix):
            return module_key
    return 'builtins'


def strip_module_prefix(module_key: str, path: str) -> str:
    """Strip the registered path prefix for *module_key* from *path*."""
    spec = catalog_module_spec(module_key)
    if spec is None:
        return path
    for prefix in spec.path_prefixes:
        if path.startswith(prefix):
            return path[len(prefix):]
    return path


def catalog_module_object(module_key: str) -> Any | None:
    """Return the loaded module object for a module key, if available."""
    for loaded_module in available_modules():
        if loaded_module.spec.key == module_key:
            return loaded_module.module
    return None


def catalog_module_spec(module_key: str) -> CatalogModuleSpec | None:
    """Return one catalog module spec by module key."""
    return _spec_by_key().get(module_key)


__all__ = [
    'CatalogModuleSpec',
    'CatalogScanSpec',
    'GLOBAL_BLACKLIST',
    'LoadedCatalogModule',
    'available_modules',
    'catalog_module_object',
    'catalog_module_spec',
    'catalog_module_specs',
    'module_options',
    'path_to_module',
    'register_catalog_module_specs',
    'resolve_attr_path',
    'runtime_global_bindings',
    'scan_targets',
    'strip_module_prefix',
]


"""Base Acherion node-definition types and shared public helpers."""

from __future__ import annotations

from abc import ABC
from collections.abc import Callable
import re
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from acherion.events import AcherionExternalEvent


_NodeParamsFactory = Callable[[str], dict[str, Any]]
_NodePinsFactory = Callable[[Any, Any], list[dict[str, str]]]
_NodeEditorRenderer = Callable[..., Any]
_NodeSourceParamIdsFactory = Callable[[Any], list[str] | tuple[str, ...]]

_IDENTIFIER_RE = re.compile(r'[^a-zA-Z0-9_]+')
_SOURCE_PARAM_SENTINEL = object()


def literal_params(value: dict[str, Any]) -> _NodeParamsFactory:
    """Return a params factory that copies one static mapping per node."""

    def _factory(_node_id: str) -> dict[str, Any]:
        return dict(value)

    return _factory


def source_param_ids(
    *param_names: str,
) -> _NodeSourceParamIdsFactory:
    """Return a factory that exposes fixed source-param ids for one node."""
    cleaned_names = tuple(
        str(name or '').strip()
        for name in param_names
        if str(name or '').strip()
    )

    def _factory(_node: Any) -> tuple[str, ...]:
        return cleaned_names

    return _factory


def acherion_node_identifier(kind: str, node: Any) -> str | None:
    """Return host-defined stable identifier for one node, when present."""
    from acherion.registry import (
        get_acherion_node_definition,
    )

    definition = get_acherion_node_definition(str(kind or '').strip())
    if definition is None:
        return None
    identifier = str(definition.node_identifier(node) or '').strip()
    return identifier or None


def acherion_auto_identifier(text: str, fallback: str = 'field') -> str:
    """Return a snake_case identifier derived from UI text."""
    identifier = _IDENTIFIER_RE.sub('_', str(text or '').strip().lower())
    identifier = re.sub(r'_+', '_', identifier).strip('_')
    if not identifier:
        identifier = fallback
    if identifier[0].isdigit():
        identifier = f'field_{identifier}'
    return identifier


def source_key_params(_node_id: str) -> dict[str, Any]:
    """Return the default source-key params payload for schema nodes."""
    return {'source_key': ''}


def custom_function_params(node_id: str) -> dict[str, Any]:
    """Return default params for a new custom-function node."""
    del node_id
    return {
        'function_path': '',
        'module': 'user',
        'arg_count': 0,
        'arg_sources': [],
    }


def pin(
    pin_id: str,
    label: str,
    pin_type: str,
    *,
    optional: bool = False,
    editor_kind: str = '',
) -> dict[str, str]:
    """Return one normalized pin specification."""
    pin = {
        'pin_id': pin_id,
        'label': label,
        'type': pin_type,
    }
    if optional:
        pin['optional'] = 'true'
    if editor_kind:
        pin['editor_kind'] = editor_kind
    return pin


_NODE_DEFINITION_OVERRIDE_FIELDS: frozenset[str] = frozenset({
    'kind',
    'label',
    'icon',
    'tooltip',
    'category',
    'flavor',
    'manual_add',
    'producer',
    'system_source',
    'system_sink',
    'exec_in',
    'exec_out',
    'default_params_factory',
    'source_param_ids_factory',
    'input_pins_factory',
    'output_pins_factory',
    'config_fields_renderer',
})


class AcherionNodeDefinition(ABC):
    """Subclassable definition object for one Acherion node type."""

    kind: str = ''
    label: str = ''
    icon: str = ''
    tooltip: str = ''
    category: str = ''
    flavor: str = ''
    manual_add: bool = False
    producer: bool = False
    system_source: bool = False
    system_sink: bool = False
    exec_in: bool = False
    exec_out: bool = False
    default_params_factory: _NodeParamsFactory | None = source_key_params
    source_param_ids_factory: _NodeSourceParamIdsFactory | None = None
    input_pins_factory: _NodePinsFactory | None = None
    output_pins_factory: _NodePinsFactory | None = None
    config_fields_renderer: _NodeEditorRenderer | None = None

    def __init__(self, **overrides: Any) -> None:
        unknown = sorted(
            key
            for key in overrides
            if key not in _NODE_DEFINITION_OVERRIDE_FIELDS
        )
        if unknown:
            unknown_fields = ', '.join(unknown)
            raise TypeError(
                f'Unknown node definition field(s): {unknown_fields}'
            )
        for key, value in overrides.items():
            setattr(self, key, value)
        self._validate_definition_metadata()

    def _definition_field(self, field_name: str) -> Any:
        if field_name in self.__dict__:
            return self.__dict__[field_name]
        return getattr(type(self), field_name)

    def _validate_definition_metadata(self) -> None:
        required_fields = ('kind', 'label', 'icon', 'category', 'flavor')
        missing = [
            field_name
            for field_name in required_fields
            if not str(self._definition_field(field_name) or '').strip()
        ]
        if missing:
            missing_text = ', '.join(missing)
            raise ValueError(
                f'Node definition is missing required field(s): '
                f'{missing_text}'
            )

    def default_title(self) -> str:
        """Return the default display title for new nodes."""
        return str(self._definition_field('label'))

    def default_params(self, *, node_id: str) -> dict[str, Any]:
        """Return initial params for a new node instance."""
        factory = self._definition_field('default_params_factory')
        if factory is None:
            params: dict[str, Any] = {}
        else:
            params = dict(factory(node_id))
        if bool(self._definition_field('exec_in')):
            params.setdefault('exec_sources', [])
        return params

    def runtime_contract(self, params: dict[str, Any]) -> dict[str, Any]:
        """Return a placeholder runtime contract for future previews."""
        return {
            'kind': str(self._definition_field('kind')),
            'params': dict(params),
            'producer': bool(self._definition_field('producer')),
            'exec_in': bool(self._definition_field('exec_in')),
            'exec_out': bool(self._definition_field('exec_out')),
        }

    def node_identifier(self, node: Any) -> str | None:
        """Return a stable identifier for one node, when applicable."""
        del node
        return None

    def schema_key_binding(self, node: Any) -> tuple[str, str] | None:
        """Return the schema namespace and param key for one node."""
        del node
        return None

    def source_param_ids(
        self,
        node: Any,
    ) -> tuple[str, ...]:
        """Return params on this node that store connected source ids."""
        factory = self._definition_field('source_param_ids_factory')
        raw_param_ids: list[str] = []
        if factory is not None:
            raw_param_ids = [
                str(param_id or '').strip()
                for param_id in factory(node)
            ]
        else:
            pins = self.input_pins(_SOURCE_PARAM_SENTINEL, node)
            if pins is not None:
                raw_param_ids = [
                    str(pin.get('pin_id') or '').strip()
                    for pin in pins
                ]

        param_ids: list[str] = []
        seen_param_ids: set[str] = set()
        for param_id in raw_param_ids:
            if not param_id or param_id in seen_param_ids:
                continue
            if param_id == 'exec_source':
                continue
            if param_id.startswith('arg:'):
                continue
            if param_id.startswith('named:'):
                continue
            if param_id.startswith('fin:'):
                continue
            seen_param_ids.add(param_id)
            param_ids.append(param_id)
        return tuple(param_ids)

    def input_pins(
        self,
        owner: Any,
        node: Any,
    ) -> list[dict[str, str]] | None:
        """Return custom input pins for one node instance, if any."""
        factory = self._definition_field('input_pins_factory')
        if factory is None:
            return None
        return [dict(pin) for pin in factory(owner, node)]

    def output_pins(
        self,
        owner: Any,
        node: Any,
    ) -> list[dict[str, str]] | None:
        """Return custom output pins for one node instance, if any."""
        factory = self._definition_field('output_pins_factory')
        if factory is None:
            return None
        return [dict(pin) for pin in factory(owner, node)]

    def render_config_fields(
        self,
        owner: Any,
        node: Any,
        *,
        refresh_editor: Callable[[], None] | None,
        apply_change: Callable[..., None],
    ) -> bool:
        """Render custom config UI for one node and return handled state."""
        renderer = self._definition_field('config_fields_renderer')
        if renderer is None:
            return False
        result = renderer(
            owner,
            node,
            refresh_editor=refresh_editor,
            apply_change=apply_change,
        )
        return result is not False

    def render_inline_controls(
        self,
        owner: Any,
        node: Any,
    ) -> bool:
        """Render compact inline node controls and return handled state."""
        del owner, node
        return False

    def external_events(
        self,
        owner: Any,
        node: Any,
    ) -> tuple[AcherionExternalEvent, ...]:
        """Return external events emitted by one node instance, if any."""
        del owner, node
        return ()


class SystemSourceNodeDefinition(AcherionNodeDefinition):
    """Base class for built-in runtime source nodes."""

    category = 'source'
    system_source = True


class UINodeDefinition(AcherionNodeDefinition):
    """Base class for UI component nodes."""

    category = 'ui'
    flavor = 'source'
    manual_add = True
    producer = True
    exec_in = True
    exec_out = True


class ComputeNodeDefinition(AcherionNodeDefinition):
    """Base class for compute nodes."""

    category = 'compute'
    flavor = 'pure'
    manual_add = True
    producer = True


class FlowNodeDefinition(AcherionNodeDefinition):
    """Base class for flow-control nodes."""

    category = 'flow'
    manual_add = True


class ObjectNodeDefinition(AcherionNodeDefinition):
    """Base class for object and method nodes."""

    category = 'object'
    flavor = 'object'
    manual_add = True


class CompositeNodeDefinition(AcherionNodeDefinition):
    """Base class for composite nodes."""

    category = 'composite'
    flavor = 'composite'


class EffectNodeDefinition(AcherionNodeDefinition):
    """Base class for effect and sink nodes."""

    category = 'effect'
    flavor = 'effect'


class SystemSinkNodeDefinition(EffectNodeDefinition):
    """Base class for built-in sink nodes."""

    system_sink = True
    exec_in = True


__all__ = [
    'AcherionNodeDefinition',
    'CompositeNodeDefinition',
    'ComputeNodeDefinition',
    'EffectNodeDefinition',
    'FlowNodeDefinition',
    'ObjectNodeDefinition',
    'SystemSinkNodeDefinition',
    'SystemSourceNodeDefinition',
    'UINodeDefinition',
    'acherion_auto_identifier',
    'acherion_node_identifier',
    'custom_function_params',
    'literal_params',
    'pin',
    'source_param_ids',
    'source_key_params',
]


"""Viewport JavaScript for the Acherion render surface."""

from __future__ import annotations

VIEWPORT_JS = """
        defaultPan(viewport) {
            const origin = this.origin(this.stage(viewport));
            const chrome = this.chrome(viewport);
            return {
                x: -origin.x + chrome.sidebarWidth + chrome.pad,
                y: -origin.y + chrome.pad,
            };
        },
        ensureViewportState(viewport) {
            if (!viewport) return;
            if (!viewport.dataset.scale) viewport.dataset.scale = '1';
            const pan = this.defaultPan(viewport);
            if (viewport.dataset.panX === undefined || viewport.dataset.panX === '')
                viewport.dataset.panX = String(pan.x);
            if (viewport.dataset.panY === undefined || viewport.dataset.panY === '')
                viewport.dataset.panY = String(pan.y);
        },
        scale(viewport) {
            this.ensureViewportState(viewport);
            const raw = parseFloat(viewport.dataset.scale || '1');
            return Number.isFinite(raw) && raw > 0 ? raw : 1;
        },
        gridBaseSize(viewport) {
            if (!viewport) return __ACH_GRID_SNAP_SIZE__;
            const styles = getComputedStyle(viewport);
            const raw = parseFloat(
                styles.getPropertyValue('--ach-grid-base-size') ||
                '__ACH_GRID_SNAP_SIZE__'
            );
            return Number.isFinite(raw) && raw > 0
                ? raw : __ACH_GRID_SNAP_SIZE__;
        },
        snapValue(viewport, value) {
            const step = this.gridBaseSize(viewport);
            return step > 0 ? Math.round(value / step) * step : value;
        },
        snapPoint(viewport, x, y) {
            return {
                x: this.snapValue(viewport, x),
                y: this.snapValue(viewport, y),
            };
        },
        updateGrid(viewport) {
            this.ensureViewportState(viewport);
            const sc = this.scale(viewport);
            const origin = this.origin(this.stage(viewport));
            const panX = parseFloat(viewport.dataset.panX || '0');
            const panY = parseFloat(viewport.dataset.panY || '0');
            const base = this.gridBaseSize(viewport);
            let size = base * sc;
            while (size < 14) size *= 2;
            const wrap = (value, step) => {
                const result = value % step;
                return result < 0 ? result + step : result;
            };
            viewport.style.setProperty('--ach-grid-size', `${size}px`);
            viewport.style.setProperty(
                '--ach-grid-offset-x',
                `${wrap(panX + origin.x * sc, size)}px`
            );
            viewport.style.setProperty(
                '--ach-grid-offset-y',
                `${wrap(panY + origin.y * sc, size)}px`
            );
        },
        applyViewportTransform(viewport) {
            this.ensureViewportState(viewport);
            this.updateGrid(viewport);
            const stage = this.stage(viewport);
            if (!stage) return;
            const sc = this.scale(viewport);
            const panX = parseFloat(viewport.dataset.panX || '0');
            const panY = parseFloat(viewport.dataset.panY || '0');
            stage.style.transform =
                `translate(${panX}px,${panY}px) scale(${sc})`;
        },
        setScaleAround(viewport, scaleValue, clientX, clientY) {
            if (!viewport) return;
            this.ensureViewportState(viewport);
            const stage = this.stage(viewport);
            if (!stage) return;
            const rect = viewport.getBoundingClientRect();
            const oldScale = this.scale(viewport);
            const nextScale = Math.max(0.08, Math.min(4.0, scaleValue));
            const panX = parseFloat(viewport.dataset.panX || '0');
            const panY = parseFloat(viewport.dataset.panY || '0');
            const worldX = (clientX - rect.left - panX) / oldScale;
            const worldY = (clientY - rect.top - panY) / oldScale;
            viewport.dataset.scale = String(nextScale);
            viewport.dataset.panX = String(
                (clientX - rect.left) - (worldX * nextScale)
            );
            viewport.dataset.panY = String(
                (clientY - rect.top) - (worldY * nextScale)
            );
            this.applyViewportTransform(viewport);
        },
        zoomViewport(viewport, factor) {
            if (!viewport) return;
            const rect = viewport.getBoundingClientRect();
            const scaleValue = this.scale(viewport) * factor;
            this.setScaleAround(
                viewport,
                scaleValue,
                rect.left + (rect.width / 2),
                rect.top + (rect.height / 2)
            );
        },
        resetViewport(viewport) {
            if (!viewport) return;
            const pan = this.defaultPan(viewport);
            viewport.dataset.scale = '1';
            viewport.dataset.panX = String(pan.x);
            viewport.dataset.panY = String(pan.y);
            this.syncViewport(viewport);
        },
        syncViewport(viewport) {
            if (!viewport) return;
            this.applyViewportTransform(viewport);
            const stage = this.stage(viewport);
            if (!stage) return;
            const update = () => this.updateConnections(stage);
            requestAnimationFrame(() => requestAnimationFrame(update));
            setTimeout(update, 0);
            setTimeout(update, 40);
            setTimeout(update, 160);
            if (document.fonts && document.fonts.ready)
                document.fonts.ready.then(update).catch(() => {});
        },
        centerGraph(viewport) {
            if (!viewport) return;
            const stage = this.stage(viewport);
            if (!stage) return;
            const origin = this.origin(stage);
            const chrome = this.chrome(viewport);
            const scaleValue = this.scale(viewport);
            const nodes = Array.from(stage.querySelectorAll('.ach-node'));
            if (!nodes.length) {
                this.resetViewport(viewport);
                return;
            }
            let minX = Infinity;
            let minY = Infinity;
            let maxX = -Infinity;
            let maxY = -Infinity;
            nodes.forEach(node => {
                const left = parseFloat(node.style.left || '0') - origin.x;
                const top = parseFloat(node.style.top || '0') - origin.y;
                minX = Math.min(minX, left);
                minY = Math.min(minY, top);
                maxX = Math.max(maxX, left + node.offsetWidth);
                maxY = Math.max(maxY, top + node.offsetHeight);
            });
            if (!Number.isFinite(minX) || !Number.isFinite(minY)) {
                this.resetViewport(viewport);
                return;
            }
            const availableWidth = Math.max(
                160,
                viewport.clientWidth - chrome.sidebarWidth - (chrome.pad * 2)
            );
            const availableHeight = Math.max(
                120,
                viewport.clientHeight - (chrome.pad * 2)
            );
            const contentWidth = Math.max(1, maxX - minX);
            const contentHeight = Math.max(1, maxY - minY);
            viewport.dataset.panX = String(
                chrome.sidebarWidth
                + chrome.pad
                + ((availableWidth - (contentWidth * scaleValue)) / 2)
                - (minX * scaleValue)
            );
            viewport.dataset.panY = String(
                chrome.pad
                + ((availableHeight - (contentHeight * scaleValue)) / 2)
                - (minY * scaleValue)
            );
            this.syncViewport(viewport);
        },
        toggleFullscreen(target) {
            if (!target) return;
            const exitFallback = () => {
                target.classList.remove('ach-workbench-local-fullscreen');
                document.body?.classList.remove(
                    'ach-workbench-fullscreen-active'
                );
            };
            const enterFallback = () => {
                target.classList.add('ach-workbench-local-fullscreen');
                document.body?.classList.add(
                    'ach-workbench-fullscreen-active'
                );
            };
            if (target.classList.contains('ach-workbench-local-fullscreen')) {
                exitFallback();
                return;
            }
            if (!document.fullscreenEnabled || !target.requestFullscreen) {
                enterFallback();
                return;
            }
            if (document.fullscreenElement === target) {
                document.exitFullscreen().catch(exitFallback);
                return;
            }
            const enter = () => {
                const fallbackTimer = setTimeout(() => {
                    if (document.fullscreenElement !== target)
                        enterFallback();
                }, 180);
                target.requestFullscreen?.().then(() => {
                    clearTimeout(fallbackTimer);
                }).catch(() => {
                    clearTimeout(fallbackTimer);
                    enterFallback();
                });
            };
            if (document.fullscreenElement) {
                document.exitFullscreen().then(enter).catch(enter);
                return;
            }
            enter();
        },
        observeAll(root) {
            (root || document).querySelectorAll('.ach-workbench').forEach(
                frame => this.syncSidebarState(frame)
            );
            (root || document).querySelectorAll('.ach-shell').forEach(
                viewport => this.observeViewport(viewport)
            );
        },
        observeViewport(viewport) {
            if (!viewport) return;
            this.ensureViewportState(viewport);
            if (viewport.__oeVlObserved) {
                this.queueSync(viewport);
                return;
            }
            viewport.__oeVlObserved = true;
            const stage = this.stage(viewport);
            if (!stage) return;
            const queue = () => this.queueSync(viewport);
            const ro = new ResizeObserver(queue);
            ro.observe(viewport);
            ro.observe(stage);
            const mo = new MutationObserver(queue);
            mo.observe(stage, {
                childList: true,
                subtree: true,
                attributes: true,
                attributeFilter: ['style', 'class'],
            });
            if (window.IntersectionObserver) {
                const io = new IntersectionObserver(
                    (entries) => {
                        if (entries.some(entry => entry.isIntersecting)) queue();
                    },
                    {threshold: [0, 0.01]}
                );
                io.observe(viewport);
                viewport.__oeVlIntersectionObserver = io;
            }
            viewport.__oeVlResizeObserver = ro;
            viewport.__oeVlMutationObserver = mo;
            queue();
        },
"""

__all__ = ['VIEWPORT_JS']


"""Connection and source-reference helpers for visual-logic graph ops."""

# pyright: reportAttributeAccessIssue=false

from __future__ import annotations

from typing import Any

from acherion.catalog import types as _catalog_types
from acherion.model import AcherionNode


class _GraphOpsConnectionsMixin:
    """Connection bookkeeping, source references, and wire actions."""

    @staticmethod
    def _normalize_exec_sources(value: object) -> list[str]:
        """Return a cleaned list of exec source ids."""
        if isinstance(value, str):
            cleaned = str(value or '').strip()
            return [cleaned] if cleaned else []
        if not isinstance(value, list):
            return []
        out: list[str] = []
        seen: set[str] = set()
        for raw_source in value:
            cleaned = str(raw_source or '').strip()
            if not cleaned or cleaned in seen:
                continue
            out.append(cleaned)
            seen.add(cleaned)
        return out

    def _exec_source_ids(self: Any, node: AcherionNode) -> list[str]:
        """Return all exec sources for one node."""
        sources = [
            self._canonical_exec_source_id(source_id)
            for source_id in self._normalize_exec_sources(
                node.params.get('exec_sources')
            )
        ]
        sources = self._normalize_exec_sources(sources)
        node.params['exec_sources'] = list(sources)
        node.params.pop('exec_source', None)
        return list(sources)

    def _canonical_exec_source_id(
        self: Any,
        source_id: str,
    ) -> str:
        """Return canonical exec source id for one stored connection."""
        cleaned = str(source_id or '').strip()
        if not cleaned or '@' in cleaned:
            return cleaned
        source_node = self._node_by_id(self._pure_node_id(cleaned))
        if source_node is None:
            return cleaned
        exec_indexes = [
            index
            for index, pin in enumerate(self._output_pin_specs(source_node))
            if str(pin.get('type') or '') == 'exec'
        ]
        if len(exec_indexes) != 1:
            return cleaned
        return f'{source_node.node_id}@{exec_indexes[0]}'

    def _set_exec_sources(
        self: Any,
        node: AcherionNode,
        source_ids: list[str],
    ) -> None:
        """Persist exec-source fan-in list."""
        normalized = [
            self._canonical_exec_source_id(source_id)
            for source_id in self._normalize_exec_sources(source_ids)
        ]
        normalized = self._normalize_exec_sources(normalized)
        node.params['exec_sources'] = list(normalized)
        node.params.pop('exec_source', None)

    def _prune_invalid_exec_connections(self: Any) -> None:
        """Drop exec links that no longer point at an exec-capable output."""
        for node in self._graph.nodes:
            source_ids = self._exec_source_ids(node)
            if not source_ids:
                continue
            has_exec_input = any(
                pin.get('pin_id') == 'exec_source'
                for pin in self._input_pin_specs(node)
            )
            if not has_exec_input:
                node.params.pop('exec_sources', None)
                continue
            valid_sources: list[str] = []
            for source_id in source_ids:
                source_node = self._node_by_id(self._pure_node_id(source_id))
                if source_node is None:
                    continue
                output_specs = self._output_pin_specs(source_node)
                output_index = self._source_pin_index(source_id)
                if output_index >= len(output_specs):
                    continue
                if str(output_specs[output_index].get('type') or '') != 'exec':
                    continue
                valid_sources.append(source_id)
            self._set_exec_sources(node, valid_sources)

    def _clear_outgoing_connections(self: Any, source_id: str) -> None:
        """Clear all outgoing wires that currently use one exact source id."""
        if not source_id:
            return
        for node in self._graph.nodes:
            for pin in self._input_pin_specs(node):
                if pin['pin_id'] == 'exec_source':
                    remaining_sources = [
                        stored_source
                        for stored_source in self._exec_source_ids(node)
                        if stored_source != source_id
                    ]
                    if len(remaining_sources) != len(
                        self._exec_source_ids(node)
                    ):
                        self._set_exec_sources(node, remaining_sources)
                    continue
                if self._input_source_id(node, pin['pin_id']) != source_id:
                    continue
                self._set_input_source(node, pin['pin_id'], '')

    def _clear_removed_source_refs(
        self: Any,
        *,
        removed_node_ids: set[str],
        removed_source_ids: set[str],
    ) -> None:
        if not removed_node_ids and not removed_source_ids:
            return
        for node in self._graph.nodes:
            for pin in self._input_pin_specs(node):
                if pin['pin_id'] == 'exec_source':
                    remaining_sources = [
                        source_id
                        for source_id in self._exec_source_ids(node)
                        if (
                            self._pure_node_id(source_id)
                            not in removed_node_ids
                            and source_id not in removed_source_ids
                        )
                    ]
                    if len(remaining_sources) != len(
                        self._exec_source_ids(node)
                    ):
                        self._set_exec_sources(node, remaining_sources)
                    continue
                source_id = self._input_source_id(node, pin['pin_id'])
                if not source_id:
                    continue
                if (
                    self._pure_node_id(source_id) in removed_node_ids
                    or source_id in removed_source_ids
                ):
                    self._set_input_source(node, pin['pin_id'], '')
        pending_source_id = str(self._pending_source_node_id or '')
        if (
            self._pure_node_id(pending_source_id) in removed_node_ids
            or pending_source_id in removed_source_ids
        ):
            self._pending_source_node_id = None

    def _input_source_id(
        self: Any,
        node: AcherionNode,
        pin_id: str,
    ) -> str:
        if pin_id == 'exec_source':
            return next(iter(self._exec_source_ids(node)), '')
        if pin_id.startswith('fin:'):
            input_node_id = pin_id.split(':', 1)[1]
            sources = dict(node.params.get('box_input_sources') or {})
            return str(sources.get(input_node_id) or '')
        if pin_id.startswith('arg:'):
            arg_index = int(pin_id.split(':', 1)[1])
            arg_sources = list(node.params.get('arg_sources') or [])
            if arg_index >= len(arg_sources):
                return ''
            return str(arg_sources[arg_index] or '')
        if pin_id.startswith('named:'):
            param_name = pin_id.split(':', 1)[1]
            named = dict(node.params.get('named_sources') or {})
            return str(named.get(param_name) or '')
        return str(node.params.get(pin_id) or '')

    def _set_input_source(
        self: Any,
        node: AcherionNode,
        pin_id: str,
        source_id: str,
    ) -> None:
        if pin_id == 'exec_source':
            self._set_exec_sources(node, [source_id] if source_id else [])
            return
        if pin_id.startswith('fin:'):
            input_node_id = pin_id.split(':', 1)[1]
            sources = dict(node.params.get('box_input_sources') or {})
            sources[input_node_id] = source_id
            node.params['box_input_sources'] = sources
            return
        if pin_id.startswith('arg:'):
            arg_index = int(pin_id.split(':', 1)[1])
            arg_sources = list(node.params.get('arg_sources') or [])
            while len(arg_sources) <= arg_index:
                arg_sources.append('')
            arg_sources[arg_index] = source_id
            node.params['arg_sources'] = arg_sources
            return
        if pin_id.startswith('named:'):
            param_name = pin_id.split(':', 1)[1]
            named = dict(node.params.get('named_sources') or {})
            named[param_name] = source_id
            node.params['named_sources'] = named
            return
        node.params[pin_id] = source_id

    def _connection_specs(self: Any) -> list[dict[str, Any]]:
        specs: list[dict[str, Any]] = []
        for node in self._graph.nodes:
            for input_index, pin in enumerate(self._input_pin_specs(node)):
                source_ids = (
                    self._exec_source_ids(node)
                    if pin['pin_id'] == 'exec_source'
                    else [self._input_source_id(node, pin['pin_id'])]
                )
                for source_id in source_ids:
                    if not source_id:
                        continue
                    source_node_id = self._pure_node_id(source_id)
                    source_node = self._node_by_id(source_node_id)
                    if source_node is None:
                        continue
                    if not self._output_pin_specs(source_node):
                        continue
                    specs.append({
                        'connection_id': self._connection_id(
                            node.node_id,
                            pin['pin_id'],
                            source_id=source_id,
                        ),
                        'source_node': source_node,
                        'full_source_id': source_id,
                        'out_pin_index': self._source_pin_index(source_id),
                        'target_node': node,
                        'pin_id': pin['pin_id'],
                        'input_index': input_index,
                    })
        return specs

    def _has_outgoing_connection(self: Any, full_source_id: str) -> bool:
        target_nid = self._pure_node_id(full_source_id)
        target_pin = self._source_pin_index(full_source_id)
        for spec in self._connection_specs():
            stored = spec['full_source_id']
            if (
                self._pure_node_id(stored) == target_nid
                and self._source_pin_index(stored) == target_pin
            ):
                return True
        return False

    def _select_connection(self: Any, connection_id: str | None) -> None:
        self._selected_connection_id = (
            str(connection_id or '').strip() or None
        )
        self.refresh()
        self._update_hint()

    def _delete_connection(self: Any, connection_id: str | None) -> None:
        data = self._split_connection_id(str(connection_id or ''))
        if data is None:
            return
        target_node_id, pin_id, source_id = data
        target_node = self._node_by_id(target_node_id)
        if target_node is None:
            return
        self._selected_connection_id = None
        if pin_id == 'exec_source':
            remaining = [
                stored_source
                for stored_source in self._exec_source_ids(target_node)
                if stored_source != source_id
            ]
            self._set_exec_sources(target_node, remaining)
        else:
            self._set_input_source(target_node, pin_id, '')
        self._notify_change()

    def _delete_selected_connection(self: Any) -> None:
        self._delete_connection(self._selected_connection_id)

    def _start_connection(self: Any, source_node_id: str) -> None:
        self._selected_connection_id = None
        if self._pending_source_node_id == source_node_id:
            self._pending_source_node_id = None
        else:
            self._pending_source_node_id = source_node_id
        self.refresh()
        self._update_hint()

    def _connect_input_pin(
        self: Any,
        node: AcherionNode,
        pin_id: str,
    ) -> None:
        if self._pending_source_node_id is None:
            self._update_hint(
                'Select an output pin first, then click an input pin.'
            )
            return
        pending_type = self._pending_output_type()
        target_type = 'any'
        for pin in self._input_pin_specs(node):
            if pin['pin_id'] == pin_id:
                target_type = str(pin.get('type') or 'any')
                break
        if pending_type != 'any' and target_type != 'any':
            if not _catalog_types.types_compatible(pending_type, target_type):
                self._update_hint(
                    f'Incompatible types: output "{pending_type}" cannot '
                    f'connect to input "{target_type}". '
                    f'Choose a compatible output pin.'
                )
                return
        if pending_type == 'exec':
            self._clear_outgoing_connections(self._pending_source_node_id)
            current_sources = self._exec_source_ids(node)
            if self._pending_source_node_id not in current_sources:
                current_sources.append(self._pending_source_node_id)
            self._set_exec_sources(node, current_sources)
        else:
            self._set_input_source(node, pin_id, self._pending_source_node_id)
        self._pending_source_node_id = None
        self._notify_change()


"""Exec-flow scope helpers for visual-logic compilation."""

from __future__ import annotations

from typing import Protocol

import acherion.node_behaviors as acherion_node_behaviors
from acherion.catalog import runtime as _catalog_runtime
from acherion.catalog import types as _catalog_types

from acherion.compiler.graph import (
    _FunctionBoxGraphView,
    _iter_param_sources,
)
from acherion.compiler.state import EmitState
from acherion.registry import (
    _template_has_exec_input,
    _template_has_exec_output,
)
from acherion.model import (
    AcherionNode,
    _node_var_name,
)


_CONSTANT_OUTPUT_TYPE_BY_VALUE_TYPE = {
    'number': 'float',
    'int': 'int',
    'text': 'str',
    'bool': 'bool',
    'dict': 'dict',
}


class _EmitNodeCallback(Protocol):
    """Protocol for node emitters used by exec-flow scope traversal."""

    def __call__(
        self,
        *,
        node: AcherionNode,
        params: dict[str, object],
        var_name: str,
        indent: str,
    ) -> None:
        ...


class _EmitScope:
    """Emit one scope by following explicit execution pins."""

    def __init__(
        self,
        scope_nodes: list[AcherionNode],
        *,
        state: EmitState,
        function_graph: _FunctionBoxGraphView,
        original_index: dict[str, int],
        entry_node_id: str = '',
        allow_implicit_roots: bool = False,
        owner_name: str | None = 'self',
    ) -> None:
        self._scope_nodes = list(scope_nodes)
        self._state = state
        self._original_index = dict(original_index)
        self._function_graph = function_graph
        self._entry_node_id = str(entry_node_id or '')
        self._allow_implicit_roots = bool(allow_implicit_roots)
        self._owner_name = owner_name
        self._node_index = {
            node.node_id: node for node in self._scope_nodes
        }
        self._emitting_nodes: set[str] = set()
        self._emitted_levels: dict[str, int] = {}
        self._emitting_exec_nodes: set[str] = set()
        self._seeding_getter_nodes: set[str] = set()
        self._dependency_stack: list[str] = []
        self._collects_seeded = False
        self._exec_targets = self._build_exec_targets()

    def seed_collects(self) -> None:
        """Initialize collect accumulators before control flow begins."""
        if self._collects_seeded:
            return
        for node in self._ordered_nodes():
            if node.kind != 'collect':
                continue
            var_name = self._var_name(node)
            self._state.lines.append(f'    {var_name} = []')
            self._state.store(node.node_id, var_name)
        self._collects_seeded = True

    def emit(self, emit_node: _EmitNodeCallback) -> None:
        """Emit this scope by following explicit exec edges."""
        self.seed_collects()
        if self._entry_node_id:
            entry_node = self._node_index.get(self._entry_node_id)
            if entry_node is not None:
                self._emit_exec_node(
                    entry_node,
                    indent_level=1,
                    emit_node=emit_node,
                )
            else:
                self._emit_entry_successors(
                    self._entry_node_id,
                    indent_level=1,
                    emit_node=emit_node,
                )
            return
        if not self._allow_implicit_roots:
            return
        for node in self._ordered_nodes():
            if not self._is_exec_root(node):
                continue
            self._emit_exec_node(node, indent_level=1, emit_node=emit_node)

    def _emit_entry_successors(
        self,
        entry_source_id: str,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        for next_node in self._exec_targets.get(entry_source_id, []):
            self._emit_exec_node(
                next_node,
                indent_level=indent_level,
                emit_node=emit_node,
            )

    def ensure_source(
        self,
        source_id: str,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        """Ensure code exists for one value source within this scope."""
        if not source_id or self._source_is_known(source_id):
            return
        node = self._node_index.get(self._pure_source_id(source_id))
        if node is None:
            return
        if not self._can_emit_as_dependency(node):
            self._seed_backward_getter_alias(
                node,
                indent_level=indent_level,
                emit_node=emit_node,
            )
            return
        self._ensure_value_emitted(
            node,
            indent_level=indent_level,
            emit_node=emit_node,
        )

    def _ordered_nodes(self) -> list[AcherionNode]:
        return sorted(
            self._scope_nodes,
            key=lambda node: self._original_index.get(node.node_id, 0),
        )

    @staticmethod
    def _normalize_exec_sources(value: object) -> list[str]:
        """Return cleaned exec-source ids from list storage."""
        if isinstance(value, str):
            cleaned = str(value or '').strip()
            return [cleaned] if cleaned else []
        if not isinstance(value, list):
            return []
        out: list[str] = []
        seen: set[str] = set()
        for raw_source in value:
            cleaned = str(raw_source or '').strip()
            if not cleaned or cleaned in seen:
                continue
            out.append(cleaned)
            seen.add(cleaned)
        return out

    def _exec_source_ids(self, node: AcherionNode) -> list[str]:
        return self._normalize_exec_sources(node.params.get('exec_sources'))

    def _build_exec_targets(self) -> dict[str, list[AcherionNode]]:
        targets: dict[str, list[AcherionNode]] = {}
        for node in self._ordered_nodes():
            for source_id in self._exec_source_ids(node):
                targets.setdefault(source_id, []).append(node)
        return targets

    def _source_is_known(self, source_id: str) -> bool:
        pure_source = self._pure_source_id(source_id)
        return (
            source_id in self._state.node_vars
            or pure_source in self._state.node_vars
        )

    def _pure_source_id(self, source_id: str) -> str:
        return source_id.split('@', 1)[0] if '@' in source_id else source_id

    @staticmethod
    def _source_pin_index(source_id: str) -> int:
        if '@' not in source_id:
            return 0
        try:
            return int(source_id.split('@', 1)[1] or 0)
        except ValueError:
            return 0

    def _var_name(self, node: AcherionNode) -> str:
        index = self._original_index[node.node_id]
        return _node_var_name(index, node)

    def _node_label(self, node_id: str) -> str:
        node = self._node_index.get(node_id)
        if node is None:
            return node_id
        title = str(node.title or '').strip()
        if title:
            return title
        return str(node.kind or node_id).strip() or node_id

    def _raise_data_cycle(self, node_id: str) -> None:
        if node_id in self._dependency_stack:
            start_index = self._dependency_stack.index(node_id)
            cycle_ids = self._dependency_stack[start_index:] + [node_id]
        elif self._dependency_stack:
            cycle_ids = self._dependency_stack + [node_id]
        else:
            cycle_ids = [node_id, node_id]
        cycle_path = ' -> '.join(
            self._node_label(cycle_node_id)
            for cycle_node_id in cycle_ids
        )
        raise ValueError(
            'Cyclic data dependency detected: '
            f'{cycle_path}. Remove one of the value wires.'
        )

    def _indent(self, indent_level: int) -> str:
        return '    ' * indent_level

    def _is_exec_root(self, node: AcherionNode) -> bool:
        if not self._exec_output_pins(node):
            return False
        if self._allow_implicit_roots:
            if (
                acherion_node_behaviors.compiler_is_backward_getter(node)
                and not self._setter_source_id(node)
            ):
                return False
            return (
                acherion_node_behaviors.compiler_is_serial_root(node)
                and not self._exec_source_ids(node)
            )
        if not _template_has_exec_input(node.kind):
            return True
        return not self._exec_source_ids(node)

    def _call_method_return_type(
        self,
        node: AcherionNode,
    ) -> str | None:
        if node.kind != 'call_method':
            return None
        method_name = str(node.params.get('method_name') or '').strip()
        if not method_name:
            return None
        class_path = self._resolve_instance_class_path(
            str(node.params.get('instance') or '').strip()
        )
        if not class_path:
            return None
        entry = _catalog_runtime.method_func_entry(class_path, method_name)
        if entry is None:
            return None
        return str(getattr(entry, 'return_type', '') or '').strip()

    def _source_output_type(
        self,
        node: AcherionNode,
        source_id: str,
    ) -> str:
        pin_index = self._source_pin_index(source_id)
        if node.kind == 'for_each':
            if pin_index == 0:
                list_source_id = str(node.params.get('list') or '').strip()
                source_node = self._node_index.get(
                    self._pure_source_id(list_source_id)
                )
                if source_node is None or source_node.node_id == node.node_id:
                    return 'any'
                list_type = self._source_output_type(source_node, list_source_id)
                item_type = _catalog_types.list_item_type_tag(list_type)
                return item_type or 'any'
            if pin_index == 1:
                return 'int'
            return 'any'
        if pin_index != 0:
            return 'any'
        if node.kind == 'constant':
            value_type = str(node.params.get('value_type') or 'number').strip()
            return _CONSTANT_OUTPUT_TYPE_BY_VALUE_TYPE.get(value_type, 'float')
        if node.kind == 'call_function':
            function_path = str(node.params.get('function_path') or '').strip()
            entry = (
                _catalog_runtime.catalog_entry(function_path)
                if function_path
                else None
            )
            if entry is None:
                return 'any'
            if bool(getattr(entry, 'is_class', False)):
                return str(getattr(entry, 'return_type', '') or 'object')
            return str(getattr(entry, 'return_type', '') or 'any')
        if node.kind == 'call_method':
            return self._call_method_return_type(node) or 'any'
        if node.kind == 'list_index':
            mode = str(node.params.get('mode') or 'index').strip()
            if mode != 'slice':
                return 'any'
            list_source_id = str(node.params.get('source') or '').strip()
            source_node = self._node_index.get(
                self._pure_source_id(list_source_id)
            )
            if source_node is None or source_node.node_id == node.node_id:
                return 'any'
            list_type = self._source_output_type(source_node, list_source_id)
            if _catalog_types.is_list_like_type_tag(list_type):
                return list_type
            return 'any'
        if node.kind == 'list_set':
            list_source_id = str(node.params.get('source') or '').strip()
            source_node = self._node_index.get(
                self._pure_source_id(list_source_id)
            )
            if source_node is None or source_node.node_id == node.node_id:
                return 'any'
            list_type = self._source_output_type(source_node, list_source_id)
            if _catalog_types.is_list_like_type_tag(list_type):
                return list_type
            return 'any'
        return _catalog_types.node_kind_to_type(node.kind)

    def _resolve_instance_class_path(
        self,
        source_id: str,
        _depth: int = 0,
    ) -> str:
        if not source_id or _depth > 5:
            return ''
        node = self._node_index.get(self._pure_source_id(source_id))
        if node is None:
            return ''
        resolved_class_path = _catalog_types.runtime_class_path_for_type_tag(
            self._source_output_type(node, source_id)
        )
        if resolved_class_path:
            return resolved_class_path
        resolved_class_path = _catalog_types.runtime_class_path_for_type_tag(
            _catalog_types.node_kind_to_type(node.kind)
        )
        if resolved_class_path:
            return resolved_class_path
        if node.kind == 'call_function':
            function_path = str(node.params.get('function_path') or '').strip()
            entry = (
                _catalog_runtime.catalog_entry(function_path)
                if function_path
                else None
            )
            if entry is not None and bool(getattr(entry, 'is_class', False)):
                return function_path
        if node.kind in {'call_method', 'get_attribute'}:
            instance_source = str(node.params.get('instance') or '').strip()
            return self._resolve_instance_class_path(
                instance_source,
                _depth + 1,
            )
        return ''

    def _call_method_has_zero_data_exec_output(
        self,
        node: AcherionNode,
    ) -> bool:
        return_type = self._call_method_return_type(node)
        if return_type is None:
            return False
        return not return_type

    def _exec_output_pins(
        self,
        node: AcherionNode,
    ) -> list[tuple[int, str]]:
        if node.kind == 'branch_route':
            return [(0, 'if_true'), (1, 'if_false')]
        if node.kind == 'for_each':
            return [(2, 'loop_body'), (3, 'completed')]
        if node.kind == 'sequencer':
            then_count = max(2, int(node.params.get('then_count', 2) or 2))
            return [
                (index, f'then:{index}')
                for index in range(then_count)
            ]
        if not _template_has_exec_output(node.kind):
            return []
        if node.kind == 'function_box':
            return [(0, 'exec')]
        has_zero_data_exec_output = (
            acherion_node_behaviors.compiler_has_zero_data_exec_output(node)
            or self._call_method_has_zero_data_exec_output(node)
        )
        exec_index = (
            0
            if has_zero_data_exec_output
            else 1
        )
        return [(exec_index, 'exec')]

    def _full_output_source_id(
        self,
        node: AcherionNode,
        pin_index: int,
    ) -> str:
        return f'{node.node_id}@{pin_index}'

    def _exec_successors(
        self,
        node: AcherionNode,
        pin_index: int,
    ) -> list[AcherionNode]:
        source_ids = [self._full_output_source_id(node, pin_index)]
        if pin_index == 0:
            # Older graphs stored pin-zero sources as plain node_id. Accept that
            # alias while new graphs use canonical node_id@pin_index everywhere.
            source_ids.append(node.node_id)
        elif pin_index > 0:
            # Mixed producer/exec nodes historically stored exec wires as plain
            # node_id. Accept that alias for backwards compatibility.
            source_ids.append(node.node_id)

        successors: list[AcherionNode] = []
        seen_node_ids: set[str] = set()
        for source_id in source_ids:
            for successor in self._exec_targets.get(source_id, []):
                if successor.node_id in seen_node_ids:
                    continue
                seen_node_ids.add(successor.node_id)
                successors.append(successor)
        return successors

    def _data_source_ids(self, node: AcherionNode) -> list[str]:
        return [
            source_id
            for source_id in _iter_param_sources(node)
            if self._pure_source_id(source_id) in self._node_index
        ]

    def _setter_source_id(self, node: AcherionNode) -> str:
        """Return the data source that makes one getter node imperative."""
        if not acherion_node_behaviors.compiler_is_backward_getter(node):
            return ''
        return str(node.params.get('source') or '').strip()

    def _seed_backward_getter_alias(
        self,
        node: AcherionNode,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        """Seed getter-only value aliases without forcing setter exec order."""
        if not acherion_node_behaviors.compiler_is_backward_getter(node):
            return
        if node.node_id in self._seeding_getter_nodes:
            self._raise_data_cycle(node.node_id)
        if self._source_is_known(node.node_id):
            return

        self._seeding_getter_nodes.add(node.node_id)
        self._dependency_stack.append(node.node_id)
        try:
            source_id = self._setter_source_id(node)
            source_expr: str | None = None
            if source_id:
                self.ensure_source(
                    source_id,
                    indent_level=indent_level,
                    emit_node=emit_node,
                )
                source_expr = self._state.source_expr(source_id)
            expr = acherion_node_behaviors.dependency_expr_for_node(
                node,
                source_expr=source_expr,
                owner_name=self._owner_name,
            )
            if expr is None:
                return
            self._state.store(node.node_id, expr)
        finally:
            self._dependency_stack.pop()
            self._seeding_getter_nodes.discard(node.node_id)

    def _ensure_dependencies(
        self,
        node: AcherionNode,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        for source_id in self._data_source_ids(node):
            self.ensure_source(
                source_id,
                indent_level=indent_level,
                emit_node=emit_node,
            )

    def _can_reemit(self, node: AcherionNode) -> bool:
        return acherion_node_behaviors.compiler_can_reemit(node)

    def _can_emit_as_dependency(self, node: AcherionNode) -> bool:
        if acherion_node_behaviors.compiler_is_backward_getter(node):
            if self._setter_source_id(node):
                return False
            return True
        if acherion_node_behaviors.compiler_is_serial_root(node):
            return True
        return not (
            _template_has_exec_input(node.kind)
            or _template_has_exec_output(node.kind)
        )

    def _ensure_value_emitted(
        self,
        node: AcherionNode,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        if node.kind in {'branch_route', 'for_each', 'function_entry'}:
            return
        if node.kind == 'sequencer':
            return
        previous_level = self._emitted_levels.get(node.node_id)
        if previous_level is not None:
            if previous_level <= indent_level:
                return
            if not self._can_reemit(node):
                return
        if node.node_id in self._emitting_nodes:
            self._raise_data_cycle(node.node_id)
        self._emitting_nodes.add(node.node_id)
        self._dependency_stack.append(node.node_id)
        try:
            self._ensure_dependencies(
                node,
                indent_level=indent_level,
                emit_node=emit_node,
            )
            emit_node(
                node=node,
                params=dict(node.params),
                var_name=self._var_name(node),
                indent=self._indent(indent_level),
            )
            self._emitted_levels[node.node_id] = indent_level
        finally:
            self._dependency_stack.pop()
            self._emitting_nodes.discard(node.node_id)

    def _emit_branch(
        self,
        node: AcherionNode,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        self._ensure_dependencies(
            node,
            indent_level=indent_level,
            emit_node=emit_node,
        )
        indent = self._indent(indent_level)
        cond_expr = self._state.source_expr(
            str(node.params.get('condition_source') or ''),
            fallback='False',
        )
        true_target = next(iter(self._exec_successors(node, 0)), None)
        false_target = next(iter(self._exec_successors(node, 1)), None)
        self._state.lines.append(f'{indent}if bool({cond_expr}):')
        if true_target is not None:
            self._emit_exec_node(
                true_target,
                indent_level=indent_level + 1,
                emit_node=emit_node,
            )
        if false_target is not None:
            self._state.lines.append(f'{indent}else:')
            self._emit_exec_node(
                false_target,
                indent_level=indent_level + 1,
                emit_node=emit_node,
            )

    def _emit_for_each(
        self,
        node: AcherionNode,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        self._ensure_dependencies(
            node,
            indent_level=indent_level,
            emit_node=emit_node,
        )
        indent = self._indent(indent_level)
        list_expr = self._state.source_expr(
            str(node.params.get('list') or ''),
            fallback='[]',
        )
        var_name = self._var_name(node)
        item_var = f'{var_name}_item'
        index_var = f'{var_name}_index'
        self._state.store(node.node_id, item_var)
        self._state.store(node.node_id, item_var, 0)
        self._state.store(node.node_id, index_var, 1)
        self._state.lines.append(
            f'{indent}for {index_var}, {item_var} in enumerate({list_expr} or []):'
        )
        body_target = next(iter(self._exec_successors(node, 2)), None)
        if body_target is not None:
            self._emit_exec_node(
                body_target,
                indent_level=indent_level + 1,
                emit_node=emit_node,
            )
        completed_target = next(iter(self._exec_successors(node, 3)), None)
        if completed_target is not None:
            self._emit_exec_node(
                completed_target,
                indent_level=indent_level,
                emit_node=emit_node,
            )

    def _emit_exec_node(
        self,
        node: AcherionNode,
        *,
        indent_level: int,
        emit_node: _EmitNodeCallback,
    ) -> None:
        if node.node_id in self._emitting_exec_nodes:
            return
        self._emitting_exec_nodes.add(node.node_id)
        try:
            if node.kind == 'function_entry':
                for pin_index, _pin_id in self._exec_output_pins(node):
                    for next_node in self._exec_successors(node, pin_index):
                        self._emit_exec_node(
                            next_node,
                            indent_level=indent_level,
                            emit_node=emit_node,
                        )
                return
            if node.kind == 'branch_route':
                self._emit_branch(
                    node,
                    indent_level=indent_level,
                    emit_node=emit_node,
                )
                return
            if node.kind == 'for_each':
                self._emit_for_each(
                    node,
                    indent_level=indent_level,
                    emit_node=emit_node,
                )
                return
            if node.kind == 'sequencer':
                for pin_index, _pin_id in self._exec_output_pins(node):
                    for successor in self._exec_successors(node, pin_index):
                        self._emit_exec_node(
                            successor,
                            indent_level=indent_level,
                            emit_node=emit_node,
                        )
                return
            self._ensure_dependencies(
                node,
                indent_level=indent_level,
                emit_node=emit_node,
            )
            emit_node(
                node=node,
                params=dict(node.params),
                var_name=self._var_name(node),
                indent=self._indent(indent_level),
            )
            for pin_index, _pin_id in self._exec_output_pins(node):
                for successor in self._exec_successors(node, pin_index):
                    self._emit_exec_node(
                        successor,
                        indent_level=indent_level,
                        emit_node=emit_node,
                    )
        finally:
            self._emitting_exec_nodes.discard(node.node_id)


"""Acherion public API for framework-agnostic graph logic."""

from __future__ import annotations

from acherion.events import (
    AcherionExternalEvent,
    RUN_EVENT_KEY,
    compose_acherion_external_events,
    default_acherion_external_events,
)
from acherion.host import (
    AcherionCodegenHost,
    AcherionConfigRenderHost,
    AcherionGraphSyncHost,
    AcherionHost,
    ComposedAcherionHost,
    compose_acherion_host,
)
from acherion.catalog.modules import (
    CatalogModuleSpec,
    CatalogScanSpec,
    register_catalog_module_specs,
)
from acherion.node import (
    AcherionNodeDefinition,
    CompositeNodeDefinition,
    ComputeNodeDefinition,
    EffectNodeDefinition,
    FlowNodeDefinition,
    ObjectNodeDefinition,
    SystemSinkNodeDefinition,
    SystemSourceNodeDefinition,
    UINodeDefinition,
    acherion_auto_identifier,
    acherion_node_identifier,
    custom_function_params,
    literal_params,
    pin,
    source_key_params,
)
from acherion.preview import (
    AcherionPreviewValueAdapter,
    AcherionPreviewRunResult,
    register_preview_value_adapter,
)
from acherion.theme import (
    AcherionThemeOverrides,
    acherion_plotly_template_payload,
    acherion_ui_colors,
    build_acherion_theme_override_css,
    build_embedded_acherion_theme_css,
    build_scoped_acherion_theme_css,
    normalize_acherion_theme_overrides,
)
from acherion.registry import (
    AcherionNodeRegistry,
    disable_acherion_node_kind,
    enable_acherion_node_kind,
    get_acherion_disabled_node_kinds,
    get_acherion_node_definition,
    get_acherion_node_definitions,
    get_acherion_enabled_node_definitions,
    get_acherion_node_registry,
    is_acherion_manual_add_kind,
    is_acherion_node_kind_enabled,
    is_acherion_producer_kind,
    is_acherion_system_sink_kind,
    is_acherion_system_source_kind,
    register_acherion_node_definition,
    set_acherion_disabled_node_kinds,
)
from acherion.validation import (
    AcherionValidationExtension,
    register_acherion_validation_extension,
)
from acherion.model import (
    AcherionGraph,
    AcherionNode,
)
from acherion.standalone_host import (
    StandaloneAcherionHost,
    compile_acherion_graph,
    execute_acherion_graph,
    load_acherion_graph_namespace,
)

__all__ = [
    'AcherionCodegenHost',
    'AcherionConfigRenderHost',
    'AcherionExternalEvent',
    'AcherionGraph',
    'AcherionGraphSyncHost',
    'AcherionHost',
    'ComposedAcherionHost',
    'AcherionNode',
    'AcherionNodeDefinition',
    'AcherionNodeRegistry',
    'AcherionPreviewValueAdapter',
    'AcherionPreviewRunResult',
    'AcherionThemeOverrides',
    'AcherionValidationExtension',
    'CatalogModuleSpec',
    'CatalogScanSpec',
    'CompositeNodeDefinition',
    'acherion_plotly_template_payload',
    'acherion_ui_colors',
    'compile_acherion_graph',
    'compose_acherion_host',
    'compose_acherion_external_events',
    'ComputeNodeDefinition',
    'disable_acherion_node_kind',
    'enable_acherion_node_kind',
    'EffectNodeDefinition',
    'execute_acherion_graph',
    'FlowNodeDefinition',
    'get_acherion_disabled_node_kinds',
    'ObjectNodeDefinition',
    'SystemSinkNodeDefinition',
    'SystemSourceNodeDefinition',
    'StandaloneAcherionHost',
    'UINodeDefinition',
    'acherion_auto_identifier',
    'acherion_node_identifier',
    'build_acherion_theme_override_css',
    'build_embedded_acherion_theme_css',
    'build_scoped_acherion_theme_css',
    'custom_function_params',
    'default_acherion_external_events',
    'get_acherion_node_definition',
    'get_acherion_node_definitions',
    'get_acherion_enabled_node_definitions',
    'get_acherion_node_registry',
    'is_acherion_manual_add_kind',
    'is_acherion_node_kind_enabled',
    'is_acherion_producer_kind',
    'is_acherion_system_sink_kind',
    'is_acherion_system_source_kind',
    'literal_params',
    'load_acherion_graph_namespace',
    'normalize_acherion_theme_overrides',
    'pin',
    'register_acherion_validation_extension',
    'register_acherion_node_definition',
    'register_catalog_module_specs',
    'register_preview_value_adapter',
    'RUN_EVENT_KEY',
    'set_acherion_disabled_node_kinds',
    'source_key_params',
]


"""Schema renderer registry and generic editor helpers for Acherion."""

from __future__ import annotations

from collections.abc import Callable, Mapping
import re
from typing import Any

from nicegui import ui

_SchemaComponentFieldRenderer = Callable[..., None]

_IDENTIFIER_RE = re.compile(r'[^a-zA-Z0-9_]+')

_SCHEMA_COMPONENT_FIELD_RENDERERS: dict[
    str,
    _SchemaComponentFieldRenderer,
] = {}


def _schema_label(node: Any, fallback: str) -> str:
    label = str(getattr(node, 'title', '') or '').strip()
    return label or fallback


def schema_label(node: Any, fallback: str) -> str:
    """Return a title-derived label with a fallback for schema metadata."""
    return _schema_label(node, fallback)


def _auto_identifier(text: str, fallback: str = 'field') -> str:
    identifier = _IDENTIFIER_RE.sub('_', str(text or '').strip().lower())
    identifier = re.sub(r'_+', '_', identifier).strip('_')
    if not identifier:
        identifier = fallback
    if identifier[0].isdigit():
        identifier = f'field_{identifier}'
    return identifier


def register_schema_component_field_renderer(
    component_kind: str,
    renderer: _SchemaComponentFieldRenderer,
) -> None:
    """Register one schema-field renderer for a component kind."""
    kind = str(component_kind or '').strip()
    if not kind:
        raise ValueError('Schema component kind must be non-empty.')
    _SCHEMA_COMPONENT_FIELD_RENDERERS[kind] = renderer


def register_schema_component_field_renderers(
    renderers: Mapping[str, _SchemaComponentFieldRenderer],
) -> None:
    """Register a batch of schema-field renderers."""
    for component_kind, renderer in renderers.items():
        register_schema_component_field_renderer(component_kind, renderer)


def _default_event_emitter(index: int) -> dict[str, str]:
    if index <= 1:
        return {
            'name': 'changed',
            'label': 'Changed',
            'description': '',
        }
    return {
        'name': f'event_{index}',
        'label': f'Event {index}',
        'description': '',
    }


def _normalise_event_emitter(
    value: Any,
    *,
    index: int,
) -> dict[str, str]:
    raw = dict(value) if isinstance(value, dict) else {}
    emitter = dict(_default_event_emitter(index))
    emitter['name'] = _auto_identifier(
        str(raw.get('name') or emitter['name']),
        emitter['name'],
    )
    emitter['label'] = str(raw.get('label') or '').strip() or emitter['label']
    emitter['description'] = str(raw.get('description') or '').strip()
    return emitter


def _normalise_event_emitters(value: Any) -> list[dict[str, str]]:
    raw_emitters = value if isinstance(value, list) else []
    emitters = [
        _normalise_event_emitter(raw_emitter, index=index + 1)
        for index, raw_emitter in enumerate(raw_emitters)
        if isinstance(raw_emitter, dict)
    ]
    seen_names: set[str] = set()
    for index, emitter in enumerate(emitters, start=1):
        base_name = _auto_identifier(
            str(emitter.get('name') or ''),
            f'event_{index}',
        )
        name = base_name
        suffix = 1
        while name in seen_names:
            name = f'{base_name}_{suffix}'
            suffix += 1
        emitter['name'] = name
        seen_names.add(name)
    return emitters


def normalise_event_emitters(value: Any) -> list[dict[str, str]]:
    """Return one normalized list of component-emitted event metadata."""
    return _normalise_event_emitters(value)


def _set_event_emitter_field(
    emitters: list[dict[str, str]],
    index: int,
    key: str,
    value: Any,
) -> list[dict[str, str]]:
    next_emitters = [dict(emitter) for emitter in emitters]
    if not (0 <= index < len(next_emitters)):
        return next_emitters
    if key == 'name':
        next_emitters[index][key] = _auto_identifier(
            str(value or '').strip(),
            f'event_{index + 1}',
        )
    else:
        next_emitters[index][key] = str(value or '').strip()
    return _normalise_event_emitters(next_emitters)


def _render_event_emitter_fields(
    *,
    state: dict[str, Any],
    set_field: Callable[..., None],
) -> None:
    emitters = _normalise_event_emitters(state.get('event_emitters'))
    ui.label('Emitted events').classes('text-xs font-semibold oe-muted')
    ui.label(
        'Optional component-scoped runtime events. Embedded hosts can map '
        'their widget callbacks into these event keys.'
    ).classes('text-xs oe-muted')
    if not emitters:
        ui.label('No emitted events yet.').classes('text-xs oe-muted')
    for index, emitter in enumerate(emitters):
        with ui.card().classes('w-full gap-2').style(
            'padding:8px; border-radius:8px'
        ):
            with ui.row().classes('w-full items-start gap-2'):
                ui.input(
                    'Event name',
                    value=str(emitter.get('name') or ''),
                    on_change=lambda e, i=index: set_field(
                        'event_emitters',
                        _set_event_emitter_field(
                            emitters,
                            i,
                            'name',
                            e.value,
                        ),
                    ),
                ).props('outlined dense').classes('flex-1 ach-editor-field')
                ui.button(
                    icon='close',
                    on_click=lambda _e, i=index: set_field(
                        'event_emitters',
                        [
                            emitter_item
                            for emitter_index, emitter_item in enumerate(emitters)
                            if emitter_index != i
                        ],
                        True,
                    ),
                ).props('flat dense round color=negative').tooltip(
                    'Remove event'
                )
            ui.input(
                'Label',
                value=str(emitter.get('label') or ''),
                on_change=lambda e, i=index: set_field(
                    'event_emitters',
                    _set_event_emitter_field(
                        emitters,
                        i,
                        'label',
                        e.value,
                    ),
                ),
            ).props('outlined dense').classes('w-full ach-editor-field')
            ui.input(
                'Description (optional)',
                value=str(emitter.get('description') or ''),
                on_change=lambda e, i=index: set_field(
                    'event_emitters',
                    _set_event_emitter_field(
                        emitters,
                        i,
                        'description',
                        e.value,
                    ),
                ),
            ).props('outlined dense').classes('w-full ach-editor-field')

    ui.button(
        'Add event',
        icon='add',
        on_click=lambda: set_field(
            'event_emitters',
            [
                *emitters,
                _default_event_emitter(len(emitters) + 1),
            ],
            True,
        ),
    ).props('flat dense')


def _render_schema_component_fields(
    *,
    component_kind: str,
    state: dict[str, Any],
    set_field: Callable[..., None],
    refresh_editor: Callable[[], None] | None,
) -> None:
    renderer = _SCHEMA_COMPONENT_FIELD_RENDERERS.get(component_kind)
    if renderer is None:
        return
    renderer(
        component_kind=component_kind,
        state=state,
        set_field=set_field,
        refresh_editor=refresh_editor,
    )


def render_schema_component_fields(
    *,
    component_kind: str,
    state: dict[str, Any],
    set_field: Callable[..., None],
    refresh_editor: Callable[[], None] | None,
) -> None:
    """Render shared editor fields for one schema-backed component."""
    _render_schema_component_fields(
        component_kind=component_kind,
        state=state,
        set_field=set_field,
        refresh_editor=refresh_editor,
    )


def render_event_emitter_fields(
    *,
    state: dict[str, Any],
    set_field: Callable[..., None],
) -> None:
    """Render shared editor fields for component-scoped emitted events."""
    _render_event_emitter_fields(
        state=state,
        set_field=set_field,
    )


__all__ = [
    'normalise_event_emitters',
    'register_schema_component_field_renderer',
    'register_schema_component_field_renderers',
    'render_event_emitter_fields',
    'render_schema_component_fields',
    'schema_label',
]